import numpy as np
from collections import OrderedDict as odict

import scipy.sparse as sparse
## "temporary" fix to deal with scipy>1.8 changing the sparse setup
try:
    from scipy.sparse import csr_array
except ImportError:
    from scipy.sparse import csr_matrix as csr_array

import so3g

from .util import get_coindices


class AxisInterface:
    """Abstract base class for axes managed by AxisManager."""

    count = None
    name = None

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        raise NotImplementedError

    def _minirepr_(self):
        return self.__repr__()

    def copy(self):
        raise NotImplementedError
    
    def rename(self, name):
        self.name = name

    def resolve(self, src, axis_index=None):
        """Perform a check or promote-and-check of this Axis against a data
        object.

        The promotion step only applies to "unset" Axes, i.e. those
        here count is None.  Not all Axis types will be able to
        support this.  Promotion involves inspection of src and
        axis_index to fix free parameters in the Axis.  If promotion
        is successful, a new ("set") Axis is returned.  If promotion
        is attempted and fails then a ValueError is raised.X

        The check step involes confirming that the data object
        described by src (and axis_index) is compatible with the
        current axis (or with the axis resulting from axis Promotion).
        Typically that simply involves confirming that
        src.shape[axis_index] == self.count.  If the check fails, a
        ValueError is raised.

        Arguments:
          src: The data object to be wrapped (e.g. a numpy array)
          axis_index: The index of the data object to test for
            compatibility.

        Returns:
          axis: Either self, or the result of promotion.

        """
        # The default implementation performs the check step.
        # Subclasses may attempt promotion, then call this.
        ref_count = src.shape[axis_index]
        if self.count != ref_count:
            raise ValueError(
                "Dimension %i of data is incompatible with axis %s" %
                (axis_index, repr(self)))
        return self

    def restriction(self, selector):
        """Apply `selector` to the elements of this axis, returning a new Axis
        of the same type and an array indexing object (a slice or an
        array of integer indices) that may be used to extract the
        corresponding elements from a vector.

        See class header for acceptable selector objects.

        Returns (new_axis, ar_index).

        """
        raise NotImplementedError

    def intersection(self, friend, return_slices=False):
        """Find the intersection of this Axis and the friend Axis, returning a
        new Axis of the same type.  Optionally, also return array
        indexing objects that select the common elements from array
        dimensions corresponding to self and friend, respectively.

        See class header for acceptable selector objects.

        Returns (new_axis), or (new_axis, ar_index_self,
        ar_index_friend) if return_slices is True.

        """
        raise NotImplementedError


class IndexAxis(AxisInterface):
    """This class manages a simple integer-indexed axis.  When
    intersecting data, the longer one will be simply truncated to
    match the shorter.  Selectors must be slice objects (with stride
    1!) or tuples to be passed into slice(), e.g. (0, 1000) or (0,
    None, 1)..

    """
    def __init__(self, name, count=None):
        super().__init__(name)
        self.count = count

    def copy(self):
        return IndexAxis(self.name, self.count)

    def __repr__(self):
        return 'IndexAxis(%s)' % self.count

    def resolve(self, src, axis_index=None):
        if self.count is None:
            return IndexAxis(self.name, src.shape[axis_index])
        return super().resolve(src, axis_index)

    def restriction(self, selector):
        if not isinstance(selector, slice):
            sl = slice(*selector)
        else:
            sl = selector
        start, stop, stride = sl.indices(self.count)
        assert stride == 1
        assert stop <= self.count
        return IndexAxis(self.name, stop - start), sl

    def __eq__(self, other):
        return self.count == other.count

    def intersection(self, friend, return_slices=False):
        count_out = min(self.count, friend.count)
        ax = IndexAxis(self.name, count_out)
        if return_slices:
            return ax, slice(count_out), slice(count_out)
        else:
            return ax


class OffsetAxis(AxisInterface):
    """This class manages an integer-indexed axis, with an accounting for
    an integer offset of any single vector relative to some absolute
    reference point.  For example, one vector could could have 100
    elements at offset 50, and a second vector could have 100 elements
    at offset -20.  On intersection, the result would have 30 elements
    at offset 50.

    The property `origin_tag` may be used to identify the absolute
    reference point.  It could be a TOD name ('obs_2020-12-01') or a
    timestamp or whatever.

    Selectors must be slice objects (with stride 1!) or tuples to be
    passed into slice(), e.g. (0, 1000) or (0, None, 1).

    """

    origin_tag = None
    offset = 0

    def __init__(self, name, count=None, offset=0, origin_tag=None):
        super().__init__(name)
        self.count = count
        self.offset = offset
        self.origin_tag = origin_tag

    def copy(self):
        return OffsetAxis(self.name, self.count, self.offset, self.origin_tag)

    def __repr__(self):
        return 'OffsetAxis(%s:%s%+i)' % (
            self.count, self.origin_tag, self.offset)

    def _minirepr_(self):
        return 'OffsetAxis(%s)' % (self.count)

    def resolve(self, src, axis_index=None):
        if self.count is None:
            return OffsetAxis(self.name, src.shape[axis_index])
        return super().resolve(src, axis_index)

    def restriction(self, selector):
        if not isinstance(selector, slice):
            sl = slice(*selector)
        else:
            sl = selector
        start, stop, stride = sl.indices(self.count + self.offset)
        assert stride == 1
        assert start >= self.offset
        assert stop <= self.offset + self.count
        return (OffsetAxis(self.name, stop - start, start, self.origin_tag),
                slice(start - self.offset, stop - self.offset, stride))

    def __eq__(self, other):
        return (self.count == other.count and
                self.offset == other.offset and
                self.origin_tag == other.origin_tag)

    def intersection(self, friend, return_slices=False):
        offset = max(self.offset, friend.offset)
        count = min(self.count + self.offset,
                    friend.count + friend.offset) - offset
        count = max(count, 0)
        ax = OffsetAxis(self.name, count, offset, self.origin_tag)
        if return_slices:
            return ax, \
                slice(offset - self.offset, count + offset - self.offset), \
                slice(offset - friend.offset, count + offset - friend.offset)
        else:
            return ax


class LabelAxis(AxisInterface):
    """This class manages a string-labeled axis, i.e., an axis where each
    element has been given a unique name.  The vector of names can be
    found in self.vals.

    Instantiation with labels that are not strings will raise a TypeError.

    On intersection of two vectors, only elements whose names appear
    in both axes will be preserved.

    Selectors should be lists (or arrays) of label strings.

    """

    def __init__(self, name, vals=None):
        super().__init__(name)
        if vals is not None:
            if len(vals):
                vals = np.array(vals)
            else:
                vals = np.array([], dtype=np.str_)
            if vals.dtype.type is not np.str_:
                raise TypeError(
                        'LabelAxis labels must be strings not %s' % vals.dtype)
        self.vals = vals

    @property
    def count(self):
        if self.vals is None:
            return None
        return len(self.vals)

    def __repr__(self):
        if self.vals is None:
            items = ['?']
        elif len(self.vals) > 20:
            items = ([repr(v) for v in self.vals[:3]] + ['...'] +
                     [repr(v) for v in self.vals[-4:]])
        else:
            items = [repr(v) for v in self.vals]
        return 'LabelAxis(%s:' % self.count + ','.join(items) + ')'

    def _minirepr_(self):
        return 'LabelAxis(%s)' % (self.count)

    def copy(self):
        return LabelAxis(self.name, self.vals)

    def resolve(self, src, axis_index=None):
        if self.count is None:
            raise RuntimeError(
                'LabelAxis cannot be naively promoted from data.')
        return super().resolve(src, axis_index)

    def restriction(self, selector):
        # Selector should be list of vals or a mask. Returns new axis and the
        # indices into self.vals that project out the elements.
        if self.vals is not None and isinstance(selector, np.ndarray) and selector.dtype == bool:
            selector = self.vals[selector]
        _, i0, i1 = get_coindices(selector, self.vals)
        assert len(i0) == len(selector)  # not a strict subset!
        return LabelAxis(self.name, selector), i1

    def __eq__(self, other):
        return (self.count == other.count and
                np.all(self.vals == other.vals))

    def intersection(self, friend, return_slices=False):
        _vals, i0, i1 = get_coindices(self.vals, friend.vals)
        ax = LabelAxis(self.name, _vals)
        if return_slices:
            return ax, i0, i1
        else:
            return ax


class AxisManager:
    """A container for numpy arrays and other multi-dimensional
    data-carrying objects (including other AxisManagers).  This object
    keeps track of which dimensions of each object are concordant, and
    allows one to slice all hosted data simultaneously.

    """

    def __init__(self, *args):
        self._axes = odict()
        self._assignments = {}  # data_name -> [ax0_name, ax1_name, ...]
        self._fields = odict()
        for a in args:
            if isinstance(a, AxisManager):
                # merge in the axes and copy in the values.
                self.merge(a)
            elif isinstance(a, AxisInterface):
                self._axes[a.name] = a.copy()
            else:
                raise ValueError("Cannot handle type %s in constructor." % a)

    @property
    def shape(self):
        return tuple([a.count for a in self._axes.values()])

    def copy(self, axes_only=False):
        out = AxisManager()
        for k, v in self._axes.items():
            out._axes[k] = v
        if axes_only:
            return out
        for k, v in self._fields.items():
            if np.isscalar(v) or v is None:
                out._fields[k] = v
            else:
                out._fields[k] = v.copy()
        for k, v in self._assignments.items():
            out._assignments[k] = v.copy()
        return out

    def _managed_ids(self):
        ids = [id(self)]
        for v in self._fields.values():
            if isinstance(v, AxisManager):
                ids.extend(v._managed_ids())
        return ids

    def __delitem__(self, name):
        if name in self._fields:
            del self._fields[name]
            del self._assignments[name]
        elif name in self._axes:
            del self._axes[name]
            for v in self._assignments.values():
                for i, n in enumerate(v):
                    if n == name:
                        v[i] = None
        else:
            raise KeyError(name)

    def move(self, name, new_name):
        """Rename or remove a data field.  To delete the field, pass
        new_name=None.

        **Example usage:**

            1. ``aman.move('hwp_angle', None)``
                Deletes the field ``hwp_angle`` from ``aman``.
            2. ``aman.move('hwp_angle', 'angle')``
                Renames the field ``hwp_angle`` to ``angle``.
            3. ``aman.move('preprocess.t2p.t2p_stats', None)``
                Deletes the field ``t2p_stats`` from the sub-AxisManager
                ``aman.preprocess.t2p``.

        """
        if name and '.' in name:
            tmp, name = name.rsplit('.', 1)
            aman = self.get(tmp)
        else:
            aman = self
        if new_name and '.' in new_name:
            tmp, new_name = new_name.rsplit('.', 1)
            new_aman = self.get(tmp)
        else:
            new_aman = self

        if new_name is None:
            del aman._fields[name]
            del aman._assignments[name]
        else:
            new_aman._fields[new_name] = aman._fields.pop(name)
            new_aman._assignments[new_name] = aman._assignments.pop(name)
        return self

    def add_axis(self, a):
        assert isinstance( a, AxisInterface)
        self._axes[a.name] = a.copy()

    def __contains__(self, name):
        attrs = name.split(".")
        tmp_item = self
        while attrs:
            attr_name = attrs.pop(0)
            if attr_name in tmp_item._fields:
                tmp_item = tmp_item._fields[attr_name]
            elif attr_name in tmp_item._axes:
                tmp_item = tmp_item._axes[attr_name]
            else:
                return False
        return True

    def __getitem__(self, name):

        # We want to support options like:
        # aman.focal_plane.xi . aman['focal_plane.xi']
        # We will safely assume that a getitem will always have '.' as the separator
        attrs = name.split(".")
        tmp_item = self
        while attrs:
            attr_name = attrs.pop(0)
            if attr_name in tmp_item._fields:
                tmp_item = tmp_item._fields[attr_name]
            elif attr_name in tmp_item._axes:
                tmp_item = tmp_item._axes[attr_name]
            else:
                raise KeyError(attr_name)
        return tmp_item

    def __setitem__(self, name, val):

        last_pos = name.rfind(".")
        val_key = name
        tmp_item = self
        if last_pos > -1:
            val_key = name[last_pos + 1:]
            attrs = name[:last_pos]
            tmp_item = self[attrs]

        if isinstance(val, AxisManager) and isinstance(tmp_item, AxisManager):
            raise ValueError("Cannot assign AxisManager to AxisManager. Please use wrap method.")

        if val_key in tmp_item._fields:
            tmp_item._fields[val_key] = val
        else:
            raise KeyError(val_key)

    def __setattr__(self, name, value):
        # Assignment to members update those members
        # We will assume that a path exists until the last member.
        # If any member prior to that does not exist a keyerror is raised.
        if "_fields" in self.__dict__ and name in self._fields.keys():
            self._fields[name] = value
        else:
            # Other assignments update this object
            self.__dict__[name] = value

    def __delattr__(self, name):
        del self._fields[name]
        del self._assignments[name]

    def __getattr__(self, name):
        # Prevent members from override special class members.
        if name.startswith("__"): raise AttributeError(name)
        try:
            val = self[name]
        except KeyError as ex:
            raise AttributeError(name) from ex
        return val

    def __dir__(self):
        return sorted(tuple(self.__dict__.keys()) + tuple(self.keys()))

    def keys(self):
        return list(self._fields.keys()) + list(self._axes.keys())

    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default

    def shape_str(self, name):
        if np.isscalar(self._fields[name]) or self._fields[name] is None:
            return ''
        s = []
        for n, ax in zip(self._fields[name].shape, self._assignments[name]):
            if ax is None:
                s.append('%i' % n)
            else:
                s.append('%s' % ax)
        return ','.join(s)

    def __repr__(self):
        def branch_marker(name):
            return '*' if isinstance(self._fields[name], AxisManager) else ''
        stuff = (['%s%s[%s]' % (k, branch_marker(k), self.shape_str(k))
                  for k in self._fields.keys()]
                 + ['%s:%s' % (k, v._minirepr_())
                    for k, v in self._axes.items()])
        return ("{}(".format(type(self).__name__)
                + ', '.join(stuff).replace('[]', '') + ")")

    @staticmethod
    def concatenate(items, axis=0, other_fields='exact'):
        """Concatenate multiple AxisManagers along the specified axis, which
        can be an integer (corresponding to the order in
        items[0]._axes) or the string name of the axis.

        This operation is difficult to sanity check so it's best to
        use it only in simple, controlled cases!  The first item is
        given significant privilege in defining what fields are
        relevant.  Fields that appear in the first item, but do not
        share the target axis, will be treated as follows depending on
        the value of other_fields:

        - If other_fields='exact' will compare entries in all items
          and if they're identical will add it. Otherwise will fail with
          a ValueError.
        - If other_fields='fail', the function will fail with a ValueError.
        - If other_fields='first', the values from the first element
          of items will be copied into the output.
        - If other_fields='drop', the fields will simply be ignored
          (and the output will only contain fields that share the
          target axis).

        """
        assert other_fields in ['exact', 'fail', 'first', 'drop']
        if not isinstance(axis, str):
            axis = list(items[0]._axes.keys())[axis]
        fields = []
        for name in items[0]._fields.keys():
            ax_dim = None
            for i, ax in enumerate(items[0]._assignments[name]):
                if ax == axis:
                    if ax_dim is not None:
                        raise ValueError('Entry %s has axis %s on more than '
                                         '1 dimension.' % (name, axis))
                    ax_dim = i
            if ax_dim is not None:
                fields.append((name, ax_dim))
        # Design the new axis.
        vals = np.hstack([item._axes[axis].vals for item in items])
        new_ax = LabelAxis(axis, vals)
        # Concatenate each entry.
        new_data = {}
        for name, ax_dim in fields:
            shape0 = None
            keepers = []
            for item in items:
                shape1 = list(item._fields[name].shape)
                if 0 in shape1:
                    continue
                shape1[ax_dim] = -1  # This dim doesn't have to match.
                if shape0 is None:
                    shape0 = shape1
                elif shape0 != shape1:
                    raise ValueError('Field %s has incompatible shapes: '
                                     % name
                                     + '%s and %s' % (shape0, shape1))
                keepers.append(item._fields[name])
            if len(keepers) == 0:
                # Well we tried.
                keepers = [items[0]._fields[name]]
            # Call class-specific concatenation if needed.
            if isinstance(keepers[0], AxisManager):
                new_data[name] = AxisManager.concatenate(
                    keepers, axis=ax_dim, other_fields=other_fields)
            elif isinstance(keepers[0], np.ndarray):
                new_data[name] = np.concatenate(keepers, axis=ax_dim)
            elif isinstance(keepers[0], csr_array):
                # Note in scipy 1.11 the default format for vstack
                # and/or hstack seems to have change, as we started
                # seeing induced cso format here.  Force preservation
                # of incoming format.
                if ax_dim == 0:
                    new_data[name] = sparse.vstack(keepers, format=keepers[0].format)
                elif ax_dim == 1:
                    new_data[name] = sparse.hstack(keepers, format=keepers[0].format)
                else:
                    raise ValueError('sparse arrays cannot concatenate along '
                                     f'axes greater than 1, received {ax_dim}')
            else:
                # The general compatible object should have a static
                # method called concatenate.
                new_data[name] = keepers[0].concatenate(keepers, axis=ax_dim)

        # Construct.
        new_axes = []
        for ax_name, ax_def in items[0]._axes.items():
            if ax_name == axis:
                ax_def = new_ax
            new_axes.append(ax_def)
        output = AxisManager(*new_axes)
        for k, v in items[0]._assignments.items():
            axis_map = [(i, n) for i, n in enumerate(v) if n is not None]
            if isinstance(items[0][k], AxisManager):
                axis_map = None  # wrap doesn't like this.
            if k in new_data:
                output.wrap(k, new_data[k], axis_map)
            else:
                if other_fields == "exact":
                    err_msg = (f"The field '{k}' does not share axis '{axis}'; " 
                              f"{k} is not identical across all items " 
                              f"pass other_fields='drop' or 'first' or else " 
                              f"remove this field from the targets.")

                    if np.any([np.isscalar(i[k]) for i in items]):
                        # At least one is a scalar...
                        if not np.all([np.isscalar(i[k]) for i in items]):
                            raise ValueError(err_msg)
                        if not np.all([_member_equal(i[k], items[0][k])
                                       for i in items[1:]]):
                            raise ValueError(err_msg)
                        output.wrap(k, items[0][k], axis_map)
                        continue

                    elif not np.all([i[k].shape==items[0][k].shape for i in items]):
                        # Has shape; shapes differ.
                        raise ValueError(err_msg)

                    elif not np.all([_member_equal(i[k], items[0][k])
                                     for i in items[1:]]):
                        # All have same shape; values not equal.
                        raise ValueError(err_msg)

                    output.wrap(k, items[0][k].copy(), axis_map)

                elif other_fields == 'fail':
                    raise ValueError(
                        f"The field '{k}' does not share axis '{axis}'; "
                        f"pass other_fields='drop' or 'first' or else "
                        f"remove this field from the targets.")
                elif other_fields == 'first':
                    # Just copy it.
                    if np.isscalar(items[0][k]):
                        output.wrap(k, items[0][k], axis_map)
                    else:
                        output.wrap(k, items[0][k].copy(), axis_map)
                elif other_fields == 'drop':
                    pass
        return output

    # Add and remove data while maintaining internal consistency.

    def wrap(self, name, data, axis_map=None,
             overwrite=False, restrict_in_place=False):
        """Add data into the AxisManager.

        Arguments:

          name (str): name of the new data.

          data: The data to register.  This must be of an acceptable
            type, i.e. a numpy array or another AxisManager.
            If scalar (or None) then data will be directly added to
            _fields with no associated axis.

          axis_map: A list that assigns dimensions of data to
            particular Axes.  Each entry in the list must be a tuple
            with the form (dim, name) or (dim, ax), where dim is the
            index of the dimension being described, name is a string
            giving the name of an axis already described in the
            present object, and ax is an AxisInterface object.
            
          overwrite (bool): If True then will write over existing data
            in field ``name`` if present.

          restrict_in_place (bool): If True, then a wrapped
            AxisManager may be modified and added, without a copy
            first.  This can be much faster, if there's no need to
            preserve the wrapped item.

        """
        if overwrite and (name in self._fields):
            self.move(name, None)
        # Don't permit AxisManager reference loops!
        if isinstance(data, AxisManager):
            assert(id(self) not in data._managed_ids())
            assert(axis_map is None)
            axis_map = [(i, v) for i, v in enumerate(data._axes.values())]
        # Handle scalars
        if np.isscalar(data) or data is None:
            if name in self._fields:
                raise ValueError(f'Key: {name} already found in {self}')
            if np.iscomplex(data):
                # Complex values aren't supported by HDF scheme right now.
                raise ValueError(f'Cannot store complex value as scalar.')
            if isinstance(data, (np.integer, np.floating, np.str_, np.bool_)):
                # Convert sneaky numpy scalars to native python int/float/str
                data = data.item()
            self._fields[name] = data
            self._assignments[name] = []
            return self
        # Promote input data to a full AxisManager, so we can call up
        # to self.merge.
        helper = AxisManager()
        assign = [None for s in data.shape]
        # Resolve each axis declaration into an axis object, and check
        # for conflict.  If an axis is passed by name only, the
        # dimensions must agree with self.  If a full axis definition
        # is included, then intersection will be performed, later.
        if axis_map is not None:
            for index, axis in axis_map:
                if not isinstance(axis, AxisInterface):
                    # So it better be a string label... that we've heard of.
                    if axis not in self._axes:
                        raise ValueError("Axis assignment refers to unknown "
                                         "axis '%s'." % axis)
                    axis = self._axes[axis]
                axis = axis.resolve(data, index)
                helper._axes[axis.name] = axis
                assign[index] = axis.name
        helper._fields[name] = data
        helper._assignments[name] = assign
        return self.merge(helper, restrict_in_place=restrict_in_place)

    def wrap_new(self, name, shape=None, cls=None, **kwargs):
        """Create a new object and wrap it, with axes mapped.  The shape can
        include axis names instead of ints, and that will cause the
        new object to be dimensioned properly and its axes mapped.

        Args:

          name (str): name of the new data.

          shape (tuple of int and std): shape in the same sense as
            numpy, except that instead of int it is allowed to pass
            the name of a managed axis.

          cls (callable): Constructor that should be used to construct
            the object; it will be called with all kwargs passed to
            this function, and with the resolved shape as described
            here.  Defaults to numpy.ndarray.

        Examples:

            Construct a 2d array and assign it the name
            'boresight_quat', with its first axis mapped to the
            AxisManager tod's "samps" axis:

            >>> tod.wrap_new('boresight_quat', shape=('samps', 4), dtype='float64')

            Create a new empty RangesMatrix, carrying a per-det, per-samp flags:

            >>> tod.wrap_new('glitch_flags', shape=('dets', 'samps'),
                             cls=so3g.proj.RangesMatrix.zeros)

        """
        if cls is None:
            cls = np.zeros
        # Turn the shape into a tuple of ints and an axis map.
        shape_ints, axis_map = [], []
        for dim, s in enumerate(shape):
            if isinstance(s, int):
                shape_ints.append(s)
            elif isinstance(s, str):
                if s in self._axes:
                    shape_ints.append(self._axes[s].count)
                    axis_map.append((dim, self._axes[s]))
                else:
                    raise ValueError(f'shape includes axis "{s}" which is '
                                     f'not in _axes: {self._axes}')
            elif isinstance(s, AxisInterface):
                # Sure, why not.
                shape_ints.append(s.count)
                axis_map.append((dim, s.copy()))
        data = cls(shape=shape_ints, **kwargs)
        return self.wrap(name, data, axis_map=axis_map)[name]

    def restrict_axes(self, axes, in_place=True):
        """Restrict this AxisManager by intersecting it with a set of Axis
        definitions.

        Arguments:
          axes (list or dict of Axis):
          in_place (bool): If in_place == True, the intersection is
            applied to self.  Otherwise, a new object is returned,
            with data copied out.

        Returns:
          The restricted AxisManager.
        """
        if in_place:
            dest = self
        else:
            dest = self.copy(axes_only=True)
            dest._assignments.update(self._assignments)
        sels = {}
        # If simple list/tuple of Axes is passed in, convert to dict
        if not isinstance(axes, dict):
            axes = {ax.name: ax for ax in axes}
        axes = {k: v for k, v in axes.items()
                if k in dest._axes and dest._axes[k] != v}
        for name, ax in axes.items():
            if name not in dest._axes:
                continue
            if dest._axes[name].count is None:
                dest._axes[name] = ax
                continue
            _, sel0, sel1 = ax.intersection(dest._axes[name], True)
            sels[name] = sel1
            dest._axes[ax.name] = ax
        for k, v in self._fields.items():
            if isinstance(v, AxisManager):
                if len(axes) == 0 and in_place:
                    dest._fields[k] = v
                else:
                    dest._fields[k] = v.restrict_axes(axes, in_place=in_place)
            elif np.isscalar(v) or v is None:
                dest._fields[k] = v
            else:
                # I.e. an ndarray.
                sslice = [sels.get(ax, slice(None))
                          for ax in dest._assignments[k]]
                sslice = tuple(dest._broadcast_selector(sslice))
                sslice = simplify_slice(sslice, v.shape)
                dest._fields[k] = v[sslice]
        return dest

    @staticmethod
    def _broadcast_selector(sslice):
        """sslice is a list of selectors, which will typically be slice(), or
        an array of indexes.  Returns a similar list of selectors, but
        with any indexing selectors promoted to a higher
        dimensionality so that the output object will be broadcast to
        the desired shape.

        For example if the input is

           (array([0,1]), slice(0,100,2), array([12,13,14]))

        then the output will be

           (array([[0],[1]]), slice(0,100,2), array([12,13,14]))

        and the result can then be used to index an array and produce
        a view with shape (2,50,3).

        """
        ex_dim = 0
        output = [s for s in sslice]
        for i in range(len(sslice) - 1, -1, -1):
            if isinstance(sslice[i], np.ndarray):
                output[i] = sslice[i].reshape(sslice[i].shape + (1,)*ex_dim)
                ex_dim += 1
        return tuple(output)

    def restrict(self, axis_name, selector, in_place=True):
        """Restrict the AxisManager by selecting a subset of items in some
        Axis.  The Axis definition and all data fields mapped to that
        axis will be modified.
        
        Arguments:
          axis_name (str): The name of the Axis.
          selector (slice or special): Selector, in a form understood
            by the underlying Axis class (see the .restriction method
            for the Axis).
          in_place (bool): If True, modifications are made to this
            object.  Otherwise, a new object with the restriction
            applied is returned.
        
        Returns:
          The AxisManager with restrictions applied.
        
        """
        if in_place:
            dest = self
        else:
            dest = self.copy(axes_only=True)
            dest._assignments.update(self._assignments)
        new_ax, sl = dest._axes[axis_name].restriction(selector)
        for k, v in self._fields.items():
            if isinstance(v, AxisManager):
                dest._fields[k] = v.copy()
                if axis_name in v._axes:
                    dest._fields[k].restrict(
                        axis_name, 
                        selector, 
                        ## copies of axes made above
                        in_place=True 
                    )
            elif np.isscalar(v) or v is None:
                dest._fields[k] = v
            else:
                sslice = [sl if n == axis_name else slice(None)
                          for n in dest._assignments[k]]
                sslice = dest._broadcast_selector(sslice)
                if in_place:
                    dest._fields[k] = v[sslice]
                else:
                    dest._fields[k] = v[sslice].copy()
        dest._axes[axis_name] = new_ax
        return dest

    @staticmethod
    def intersection_info(*items):
        """Given a list of AxisManagers, scan the axes and combine (intersect)
        any common axes.  Returns a dict that maps axis name to
        restricted Axis object.

        """
        # Get the strictest intersection of each axis.
        axes_out = odict()
        for aman in items:
            for ax in aman._axes.values():
                if ax.count is None:
                    continue
                if ax.name not in axes_out:
                    axes_out[ax.name] = ax.copy()
                elif axes_out[ax.name] != ax:
                    axes_out[ax.name] = axes_out[ax.name].intersection(
                        ax, False)
        return axes_out

    def merge(self, *amans, restrict_in_place=False):
        """Merge the data from other AxisMangers into this one.  Axes with the
        same name will be intersected.

        If restrict_in_place=True, then the amans may be modified as
        they are added to the output objcet.  When that arg is False,
        the incoming amans are all copied, even if no modifications
        are needed.

        """
        # Before messing with anything, check for key interference.
        fields = set(self._fields.keys())
        for aman in amans:
            newf = set(aman._fields.keys())
            both = fields.intersection(newf)
            if len(both):
                raise ValueError(f'Key conflict: more than one merge target '
                                 f'shares keys: {both}')
            fields.update(newf)

        # Get the intersected axis descriptions.
        axes_out = self.intersection_info(self, *amans)
        # Reduce the data in self, update our axes.
        self.restrict_axes(axes_out)
        # Import the other ones.
        for aman in amans:
            aman = aman.restrict_axes(axes_out, in_place=restrict_in_place)
            for k, v in aman._axes.items():
                if k not in self._axes:
                    self._axes[k] = v
            for k, v in aman._fields.items():
                assert(k not in self._fields)  # Should have been caught in pre-check
                self._fields[k] = v
            self._assignments.update(aman._assignments)
        return self

    def save(self, dest, group=None, overwrite=False, compression=None,
             encodings=None):
        """Write this AxisManager data to an HDF5 group.  This is an
        experimental feature primarily intended to assist with
        debugging.  The schema is subject to change, and it's possible
        that not all objects supported by AxisManager can be
        serialized.

        Args:
          dest (str or h5py.Group): Place to save it (in combination
            with group).
          group (str or None): Group within the HDF5 file (relative to
            dest).
          overwrite (bool): If True, remove any existing thing at the
            specified address before writing there.
          compression (str or None): Compression filter to apply. E.g.
            'gzip'. This string is passed directly to HDF5 dataset
            routines.
          encodings (dict or None): Special instructions for encoding
            / compression. See notes.

        Notes:
          If dest is a string, it is taken to be an HDF5 filename and
          is opened in 'a' mode.  The group, in that case, is the
          full group name in the file where the data should be written.

          If dest is an h5py.Group, the group is the group name in the
          file relative to dest.

          The overwrite argument only matters if group is passed as a
          string.  A RuntimeError is raised if the group address
          already exists and overwrite==False.

          For example, these are equivalent::

            # Filename + group address:
            axisman.save('test.h5', 'x/y/z')

            # Open h5py.File + group address:
            with h5py.File('test.h5', 'a') as h:
              axisman.save(h, 'x/y/z')

            # Partial group address
            with h5py.File('test.h5', 'a') as h:
              g = h.create_group('x/y')
              axisman.save(g, 'z')

          When passing a filename, the code probably won't use a
          context manager... so if you want that protection, open your
          own h5py.File as in the 2nd and 3rd example.

          The encodings dict is used to support flacarray compression
          of fields.  Suppose member 'quant_field' should be
          FLAC-compressed, with precision 0.1.  Then pass::

            encodings={
              'quant_field': {
                'type': 'flacarray':
                'args': {
                  'quanta': 0.1
                }
              }
            }

          For fields in sub-axismanagers, nest the specification; e.g.::

            encodings={
              'subaman': {
                 'quant_field': ...
              }
            }

          Only multi-dimensional int and float arrays may be so
          compressed.  Unconsumed encoding information
          (e.g. specifying compression for a non-existent field or a
          field that is not an array) will cause an exception to be
          raised.

        """
        from .axisman_io import _save_axisman
        return _save_axisman(self, dest, group=group, overwrite=overwrite,
                             compression=compression, encodings=encodings)

    @classmethod
    def load(cls, src, group=None, fields=None):
        """Load a saved AxisManager from an HDF5 file and return it.  See docs
        for save() function.

        The (src, group) args are combined in the same way as (dest,
        group) in the save function.  Examples::

          axisman = AxisManager.load('test.h5', 'x/y/z')

          with h5py.File('test.h5', 'r') as h:
            axisman = AxisManager.load(h, 'x/y/z')

        If the fields argument is specified, it must be a list of
        strings indicating what subfields of the stored AxisManager
        should be extracted.  For nested entries, connect fields with
        ".".  For example ``fields=['subaman.field1',
        'subaman.field2']``.  When fields is specified, _all_ axes
        from the AxisManager are included in the result, even if not
        directly referenced by the requested fields; this behavior is
        subject to change.

        """
        from .axisman_io import _load_axisman
        return _load_axisman(src, group, cls, fields=fields)


def simplify_slice(sslice, shape):
    """Given a tuple of slices, such as what __getitem__ might produce, and the
    shape of the array it would be applied to, return a new tuple of slices that
    accomplices the same thing, but while avoiding costly general slices if possible."""
    res = []
    for n, s in zip(shape, sslice):
        # Numpy arrays slicing is expensive, and unnecessary if they just select
        # the same elemnts in the same order
        if isinstance(s, np.ndarray):
            # Is this a trivial numpy slice? If so, replace it
            if s.size == n and np.all(s == np.arange(n)):
                res.append(slice(None))
            # Otherwise bail, and keep the whole original
            else:
                return sslice
        # For anything else just pass it through. This includes normal slices
        else: res.append(s)
    return tuple(res)


def _member_equal(a, b):
    """Check equality of two things you might find in an AxisManager.

    """
    if isinstance(a, np.ndarray) or np.isscalar(a):
        # Use the array_equal test for scalars because it's nan-smart.
        equal_nan = (np.asarray(a).dtype.char not in 'SUO')
        return np.array_equal(a, b, equal_nan=equal_nan)
    elif isinstance(a, so3g.RangesInt32):
        return np.array_equal(a.ranges(), b.ranges()) and a.count == b.count
    elif isinstance(a, so3g.proj.RangesMatrix):
        return all([_member_equal(_a, _b) for _a, _b in zip(a, b)])
    else:
        return a == b



import unittest
import os
import tempfile


from sotodlib.site_pipeline import jobdb

class TestBasic(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self.tempdir.cleanup()

    def test_00_smoke(self):
        # Create some jobs
        jdb = jobdb.JobManager(sqlite_file=':memory:')

        jdb.create_job('jclass1', {'obs_id': '123455'})
        jdb.create_job('jclass1', {'obs_id': '123456'}, jstate='done')
        jdb.create_job('jclass1', {'obs_id': '123457'})
        with self.assertRaises(jobdb.JobNotUniqueError):
            jdb.create_job('jclass1', {'obs_id': '123456'})
        jdb.create_job('jclass2', {'obs_id': '123456'})

        # Counting
        jobs_to_do = jdb.get_jobs(jclass='jclass1', jstate='open')
        self.assertEqual(len(jobs_to_do), 2)

        # Locking
        j = jobs_to_do[0]
        job = jdb.lock(j.id)
        with self.assertRaises(jobdb.JobLockedError):
            job = jdb.lock(j.id)
        jdb.unlock(job)
        
        job = jdb.lock(j.id)
        jdb.unlock(job.id, merge=False)

        # State write-back
        for row in jobs_to_do:
            print(f'Finishing {row.id} ...')
            with jdb.locked(row) as job:
                job.jstate = 'done'

        jobs_to_do = jdb.get_jobs(jclass='jclass1', jstate='open')
        self.assertEqual(len(jobs_to_do), 0)

        self.assertNotEqual(len(jdb.get_jobs(jclass='jclass1', jstate='all')), 0)

        # Deleting
        jobs_to_delete = jdb.get_jobs(jclass='jclass1', jstate='done')
        for j in jobs_to_delete:
            jdb.remove_job(j.id)

        self.assertEqual(len(jdb.get_jobs(jclass='jclass1', jstate='all')), 0)

        # Create-and-operate
        j = jdb.create_job('jlcass2', {'obs_id': '123455'})
        with jdb.locked(j) as job:
            job.mark_visited()

    def test_10_report(self):
        db_file = os.path.join(self.tempdir.name, 'test_10.sqlite')
        jdb = jobdb.JobManager(sqlite_file=db_file)

        jdb.create_job('jclass1', {'obs_id': '123455'})
        jdb.create_job('jclass1', {'obs_id': '123456'}, jstate='done')
        jdb.create_job('jclass1', {'obs_id': '123457'})
        with self.assertRaises(jobdb.JobNotUniqueError):
            jdb.create_job('jclass1', {'obs_id': '123456'})
        jdb.create_job('jclass2', {'obs_id': '123456'})

        print()
        jobdb.cli(['--sqlite-file', db_file, 'select'])

    def test_20_locks(self):
        db_file = os.path.join(self.tempdir.name, 'test_20.sqlite')
        jdb = jobdb.JobManager(sqlite_file=db_file)

        jdb.create_job('jclass1', {'obs_id': '123455'})
        jdb.create_job('jclass1', {'obs_id': '123456'})
        jdb.create_job('jclass1', {'obs_id': '123457'})

        jobs = jdb.get_jobs(jclass='jclass1', jstate='open')
        with jdb.locked(jobs[0].id):
            with self.assertRaises(jobdb.JobLockedError):
                jdb.lock(jobs[0].id)
            with jdb.locked(jobs[1].id):
                pass
            jdb.clear_locks('all')
            jdb.lock(jobs[0].id)

        with self.assertRaises(jobdb.JobNotOwnedError):
            with jdb.locked(jobs, count=10) as jobs:
                # Simulate another entity stealing a lock.
                jx = jdb.lock(jobs[1].id, owner='xyz', force=True)

    def test_30_resource(self):
        jdb = jobdb.JobManager(sqlite_file=':memory:')

        kls = 'resource1'
        jdb.create_job(kls, {'channel': 'c1'})
        jdb.create_job(kls, {'channel': 'c2'})
        jdb.create_job(kls, {'channel': 'c3'})
        jdb.create_job(kls, {'channel': 'c4'})

        r = jdb.get_resource(kls)
        assert r is not None
        del r

        rs = []
        for i in range(5):
            r = jdb.get_resource(kls)
            assert (r is not None) ^ (i >= 4)
            rs.append(r)
        del rs
        #jdb.clear
        rs1 = jdb.get_resource(kls, n=3)
        assert len(rs1) == 3
        rs2 = jdb.get_resource(kls, n=3)
        assert len(rs2) == 1
        del rs1, rs2

        rs = jdb.get_resource(kls, tags={'channel': 'c2'}, n=4)
        assert len(rs) == 1



import datetime as dt, os.path as op, os
import numpy as np
from collections import OrderedDict
from typing import List
import yaml, traceback
import shutil
import logging
from pathlib import Path
from glob import glob
import time

import sqlalchemy as db
from sqlalchemy import or_, and_, not_
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
import so3g
from spt3g import core

from .bookbinder import BookBinder, TimeCodeBinder
from .load_smurf import (
    G3tSmurf,
    Observations as G3tObservations,
    SmurfStatus,
    get_channel_info,
    TimeCodes,
    SupRsyncType,
    Files,
)
from .datapkg_utils import load_configs, get_imprinter_config
from .check_book import BookScanner
from .g3thk_db import G3tHk, HKFiles
from ..site_pipeline.util import init_logger


####################
# useful constants #
####################

# book status
WONT_BIND = -2 # books we have tagged by hand that we will not bind
FAILED = -1 # books where automated binding failed
UNBOUND = 0
REBIND = 1
BOUND = 2
UPLOADED = 3
DONE = 4

# tel tube, stream_id, slot mapping
VALID_OBSTYPES = ["obs", "oper", "smurf", "hk", "stray", "misc"]

# file patterns excluded from smurf books
SMURF_EXCLUDE_PATTERNS = ["*.dat", "*_mask.txt", "*_freq.txt"]

# file size limits
MAX_OBS_LVL2_SIZE = 10 # Gb per level 2 file
MAX_OPER_LVL2_SIZE = 5 # Gb per level 2 file

class BookExistsError(Exception):
    """Exception raised when a book already exists in the database"""
    pass

class BookBoundError(Exception):
    """Exception raised when a book is already bound"""
    pass

class NoFilesError(Exception):
    """Exception raised when no files are found in the book"""
    pass

class MissingReadoutIDError(Exception):
    """Exception raised when we find books with observations with missing readout IDs"""
    pass

class OverlapObsError(Exception):
    """Exception raised when we find observations that could be registered to
    multiple books"""
    pass

class FileTooLargeError(Exception):
    """Exception raised when we find level 2 files that are larger than our
    maximum allowable sizes"""
    pass

###################
# database schema #
###################

Base = declarative_base()


class Observations(Base):
    """
    Attributes
    ----------
    obs_id: observation id
    bid: book id

    """

    __tablename__ = "observations"
    obs_id = db.Column(db.String, primary_key=True)
    bid = db.Column(db.String, db.ForeignKey("books.bid"))
    book = relationship("Books", back_populates="obs")


class Books(Base):
    """
    Attributes
    ----------
    bid: book id
    start: start time of book
    stop: stop time of book
    max_channels: maximum number of channels in book
    obs: list of observations within the book
    type: type of book, e.g., "oper", "hk", "obs"
    status: integer, stage of processing, 0 is unbound, 1 is bound,
    message: error message if book failed
    tel_tube: telescope tube
    slots: slots in comma separated string
    created_at: time when book was created
    updated_at: time when book was updated
    timing: bool, whether timing system is on
    path: str, location of book directory relative to output_root of imprinter
    lvl2_deleted: bool, if level 2 data has been purged yet
    """

    __tablename__ = "books"
    bid = db.Column(db.String, primary_key=True)
    start = db.Column(db.DateTime)
    stop = db.Column(db.DateTime)
    max_channels = db.Column(db.Integer)
    obs = relationship("Observations", back_populates="book")  # one to many
    type = db.Column(db.String)
    status = db.Column(db.Integer, default=UNBOUND)
    message = db.Column(db.String, default="")
    tel_tube = db.Column(db.String)
    slots = db.Column(db.String)
    created_at = db.Column(db.DateTime, default=dt.datetime.utcnow)
    updated_at = db.Column(
        db.DateTime, default=dt.datetime.utcnow, onupdate=dt.datetime.utcnow
    )
    timing = db.Column(db.Boolean)
    path = db.Column(db.String)
    lvl2_deleted = db.Column(db.Boolean, default=False)
    schema = db.Column(db.Integer, default=0)

    def __repr__(self):
        return f"<Book: {self.bid}>"


##############
# main logic #
##############


# convenient decorator to repeat a method over all data sources
def loop_over_tubes(method):
    def wrapper(self, *args, **kwargs):
        outs = []
        for tube in self.tubes:
            x = method(self, tube, *args, **kwargs)
            if x is not None:
                outs.extend(x)
        if len(outs)>0:
            return outs        
    return wrapper


class Imprinter:
    def __init__(self, im_config=None, db_args={}, logger=None, make_db=False):
        """Imprinter manages the book database.
        
        Imprinter at the site is set up to be one per level 2 daq node with a
        one-to-one matching between imprinter instance and g3tsmurf instance. 

        Example configuration file::
        
          db_path: imprinter.db
          daq_node: daq-node
          g3tsmurf: /path/to/config.yaml
          output_root: /abs/path/to/output
          librarian_conn: string (optional)
          build_hk: True 
          build_det: True
          hk_fields:
            az: acu.acu_udp_stream.Corrected_Azimuth
            el: acu.acu_udp_stream.Corrected_Elevation
            boresight: acu.acu_udp_stream.Corrected_Boresight
            az_mode:  acu.acu_status.Azimuth_mode
            hwp_freq: hwp-bbb-e1.HWPEncoder.approx_hwp_freq

          tel_tubes:
            tel_tube1:
              tube_slot: c0
              wafer_slots:
                - wafer_slot: ws0
                  stream_id: stream_id0
                - wafer_slot: ws1
                  stream_id: stream_id1
                - wafer_slot: ws2
                  stream_id: stream_id2
            tel_tube2:
              tube_slot: i1
              wafer_slots:
                - wafer_slot: ws0
                  stream_id: stream_id3
                - wafer_slot: ws1
                  stream_id: stream_id4
                - wafer_slot: ws2
                  stream_id: None
          
        Standard Book directory structure based off config file example:
        output_root/
            tel_tube1/
                obs/
                oper/
            tel_tube2/
                obs/
                oper/
            daq-node/
                smurf/
                stray/
                hk/
        
        The tel_tubes entry lists the different tel-tubes that will be bound
        into separate books. Each tel_tube entry is expected to have a tube_slot
        name and is required to have a wafer_slots list. Each wafer slot must
        have a wafer_slot and stream_id entry. If stream_id is None then it will
        be treated as an empty slot. The wafers/stream_ids listed in each
        tel_tube are bound together. For the SATs we expect one entry under 
        tel_tubes but the LAT will have many.
        
        The build_det_data and build_hk entries determines if detector books
        (obs,oper,smurf,stray) and housekeeping (hk) books should be made,
        respectively. 

        Parameters
        ----------
        im_config: str
            path to imprinter configuration file
        db_args: dict
            arguments to pass to sqlalchemy.create_engine
        logger: logger
            logger object
        """

        # load config file and parameters
        self.config = load_configs(im_config)

        self.db_path = self.config.get("db_path")
        self.daq_node = self.config.get("daq_node")        
        self.output_root = self.config.get("output_root")
        self.g3tsmurf_config = self.config.get("g3tsmurf")
        self.require_hwp = self.config.get("require_hwp", True)
        g3tsmurf_cfg = load_configs(self.g3tsmurf_config)
        self.lvl2_data_root = g3tsmurf_cfg["data_prefix"]

        self.build_hk = self.config.get("build_hk")
        self.build_det = self.config.get("build_det")

        self.logger = logger
        if logger is None:
            self.logger = logging.getLogger("imprinter")
            if not self.logger.hasHandlers():
                self.logger = init_logger("imprinter")

        self.tube_configs = self.config.get("tel_tubes", {})
        self.tubes = dict()
        for tube, cfg in self.tube_configs.items():
            self.tubes[tube] = dict()
            self.tubes[tube]['slots'] = list()
            if 'wafer_slots' not in self.tube_configs[tube]:
                if 'slots' in self.tube_configs[tube]:
                    self.tubes[tube]['slots'] = self.tube_configs[tube]['slots']
                    self.logger.warning(f"Tube {tube} missing wafer_slots "
                    "entry. Are you using an old configuration format?")
                    continue
                
            for slot in self.tube_configs[tube]['wafer_slots']:
                if slot['stream_id'].lower() == 'none':
                    self.tubes[tube]['slots'].append(
                        f"NONE_{slot['wafer_slot']}"
                    )
                else:
                    self.tubes[tube]['slots'].append(slot['stream_id'])
                                        
        # raise error if database does not exist
        if not op.exists(self.db_path):
            if not make_db:
                raise ValueError(
                    f"Imprinter database path {self.db_path} does not exist "
                    "pass make_db=True to make database"
                )
            else:
                # check whether db_path directory exists
                if not op.exists(op.dirname(self.db_path)):
                    # we first make sure that the folder exists
                    os.makedirs(
                        op.abspath(op.dirname(self.db_path)), exist_ok=True
                    )
        self.engine = db.create_engine(f"sqlite:///{self.db_path}", **db_args)

        # create all tables or (do nothing if tables exist)
        Base.metadata.create_all(self.engine)

        self.session = None
        self.g3tsmurf_session = None
        self.archive = None
        self.hk_archive = None
        self.librarian = None

    @classmethod
    def for_platform(cls, platform, *args, **kwargs):
        config = get_imprinter_config(platform)
        return cls( config, *args, **kwargs)

    def get_session(self):
        """Get a new session or return the existing one

        Returns
        -------
        session: BookDB session

        """
        if self.session is None:
            Session = sessionmaker(bind=self.engine)
            self.session = Session()
        return self.session

    def get_g3tsmurf_session(self, return_archive=False):
        """Get a new g3tsmurf session or return an existing one.

        Parameter
        ---------
        source: str
            data source, e.g., "sat1", "tsat", "latrt"
        return_archive: bool
            whether to return SMURF archive object

        Returns
        -------
        session: g3tsmurf session
        archive: if return_archive is true

        """
        if self.g3tsmurf_session is None:
            (
                self.g3tsmurf_session,
                self.archive,
            ) = create_g3tsmurf_session(self.g3tsmurf_config)
        if not return_archive:
            return self.g3tsmurf_session
        return self.g3tsmurf_session, self.archive

    def get_g3thk(self):
        """Get a G3tHk database using the g3tsmurf config file."""
        if self.hk_archive is None:
            self.hk_archive = G3tHk.from_configs(
                self.g3tsmurf_config
            )
        return self.hk_archive

    def register_book(self, obsset, bid=None, commit=True, session=None):
        """Register book to database

        Parameters
        ----------
        obsset: ObsSet object
            thin wrapper of a list of observations
        bid: str
            book id
        commit: boolean
            if True, commit the session
        session: SQLAlchemy session
            BookDB session

        Returns
        -------
        book: book object

        """
        if session is None:
            session = self.get_session()
        # create book id if not provided
        if bid is None:
            bid = obsset.get_id()
        assert obsset.mode is not None
        assert obsset.mode in ['obs','oper']
        # check whether book exists in the database
        if self.book_exists(bid, session=session):
            raise BookExistsError(f"Book {bid} already exists in the database")
        self.logger.info(f"Registering book {obsset} (mode: {obsset.mode})")

        # fill in book attributes: start, end, max_channels
        # start and end are the earliest and latest observation time
        # max_channels is the maximum number of channels in the book
        if any([len(o.files) == 0 for o in obsset]):
            raise NoFilesError(f"No files found for observations in {bid}")
        # get start and end times (initial guess)
        start_t = np.max([np.min([f.start for f in o.files]) for o in obsset])
        stop_t = np.min([np.max([f.stop for f in o.files]) for o in obsset])
        max_channels = int(
            np.max([np.max([f.n_channels for f in o.files]) for o in obsset])
        )
        timing_on = np.all([o.timing for o in obsset])

        # create observation and book objects
        observations = [
            Observations(obs_id=obs_id) for obs_id in obsset.obs_ids
        ]
        book = Books(
            bid=bid,
            obs=observations,
            type=obsset.mode,
            status=UNBOUND,
            start=start_t,
            stop=stop_t,
            max_channels=max_channels,
            tel_tube=obsset.tel_tube,
            slots=",".join(
                [s for s in obsset.slots if obsset.contains_stream(s)]
            ),  # not worth having a extra table
            timing=timing_on,
            schema=0,
        )
        book.path = self.get_book_path(book)

        self.logger.info(f"registering {bid}")
        # add book to database
        session.add(book)
        if commit:
            session.commit()
        return book

    def register_books(self, bids, obs_lists, commit=True, session=None):
        """Register multiple books to database.

        Parameters
        ----------
        bids: list
            a list of book ids (str)
        obs_lists: list
            a list of list where the smaller list consists of
            different observations and the outer loop goes through
            different books
        commit: boolean
            if True, commit the session
        session: BookDB session

        """
        if session is None:
            session = self.get_session()
        for bid, obs_list in zip(bids, obs_lists):
            self.register_book(session, bid, obs_list, commit=False)
        if commit:
            session.commit()

    def register_hk_books(self, min_ctime=None, max_ctime=None, session=None):
        """Register housekeeping books to database"""
        session = session or self.get_session()
        if not self.build_hk:
            return

        if min_ctime is None:
            min_ctime = 16000e5
        if max_ctime is None:
            max_ctime = 5e10

        # all ctime dir except the last ctime dir will be considered complete
        ctime_dirs = sorted(glob(op.join(self.lvl2_data_root, "hk", "*")))
        for ctime_dir in ctime_dirs[:-1]:
            ctime = op.basename(ctime_dir)
            if int(ctime) < int(min_ctime//1e5):
                continue
            if int(ctime) > int(max_ctime//1e5):
                continue
            book_id = f"hk_{ctime}_{self.daq_node}"
            # check whether the book exists
            if self.get_book(book_id) is not None:
                continue
            self.logger.info(f"registering {book_id}")
            book = Books(
                bid=book_id,
                type="hk",
                status=UNBOUND,
                start=dt.datetime.utcfromtimestamp(int(ctime) * 1e5),
                stop=dt.datetime.utcfromtimestamp((int(ctime) + 1) * 1e5),
                tel_tube=self.daq_node,  
                schema=0,
            )
            book.path = self.get_book_path(book)
            session.add(book)
            session.commit()

    def register_timecode_books(
        self, 
        min_ctime=None, 
        max_ctime=None,
        session=None
    ):
        """Register smurf and stray books with database. We do not expect many
        stray books as nearly all .g3 files will be associated with some obs or
        oper book.

        stray and smurf books are made per DAQ node, meaning there will be one
        per imprinter instance on the site setup. The making of these books
        depends on the imprinter configuration having the correct stream_ids
        indexed per tel-tube. All stream_ids not part of those lists
        will be moved to stray books but only the finalization information from
        the specified stream ids will be used to determine when to output these
        books.

        smurf books are registered whenever all the relevant metadata timecode
        entries have been found. stray books are registered when metadata and
        file timecode entries exist AND all obs/oper books in that time
        range have been bound successfully.
        """

        if not self.build_det:
            return
        session = session or self.get_session()
        g3session, SMURF = self.get_g3tsmurf_session(return_archive=True)

        final_time = SMURF.get_final_time(
            self.all_slots, min_ctime, max_ctime, check_control=False
        )
        final_tc = int(final_time//1e5)
        servers = SMURF.finalize["servers"]
        meta_agents = [s["smurf-suprsync"] for s in servers]
        files_agents = [s["timestream-suprsync"] for s in servers]

        meta_query = or_(*[TimeCodes.agent == a for a in meta_agents])
        files_query = or_(*[TimeCodes.agent == a for a in files_agents])

        tcs = g3session.query(TimeCodes.timecode)
        if min_ctime is not None:
            tcs = tcs.filter(TimeCodes.timecode >= int(min_ctime//1e5))
        if max_ctime is not None:
            tcs = tcs.filter(TimeCodes.timecode <= int(max_ctime//1e5)+1)
        tcs = tcs.distinct().all()

        for (tc,) in tcs:
            if tc >= final_tc:
                self.logger.info(
                    f"Not ready to make timecode books for {tc} because final"
                    f" timecode is {final_tc}"
                )
                continue
            q = g3session.query(TimeCodes).filter(
                TimeCodes.timecode == tc,
            )
            files = g3session.query(TimeCodes.agent).filter(
                TimeCodes.timecode==tc,
                files_query,
                TimeCodes.suprsync_type == SupRsyncType.FILES.value,
            ).distinct().all()
            meta = g3session.query(TimeCodes.agent).filter(
                TimeCodes.timecode==tc,
                meta_query,
                TimeCodes.suprsync_type == SupRsyncType.META.value,
            ).distinct().all()

            if not len(meta) == len(meta_agents):
                # not ready to make any books
                self.logger.debug(f"Not ready to make timecode books for {tc}")
                continue
            
            ## register smurf  books
            book_start = dt.datetime.utcfromtimestamp(tc * 1e5)
            book_stop = dt.datetime.utcfromtimestamp((tc + 1) * 1e5)

            book_id = f"smurf_{tc}_{self.daq_node}"
            if self.get_book(book_id) is None:
                self.logger.info(f"registering {book_id}")
                smurf_book = Books(
                    bid=book_id,
                    type="smurf",
                    status=UNBOUND,
                    tel_tube=self.daq_node,
                    start=book_start,
                    stop=book_stop,
                    schema=1,
                )
                smurf_book.path = self.get_book_path(smurf_book)
                session.add(smurf_book)
                session.commit()

            if not len(files) == len(files_agents):
                #not ready to make stray books
                self.logger.debug(f"Not ready to make stray books for {tc}")
                continue

            book_id = f"stray_{tc}_{self.daq_node}"
            if self.get_book(book_id) is not None:
                # book already registered
                continue

            # look for failed or unbound books
            q = session.query(Books).filter(
                Books.start >= book_start,
                Books.start < book_stop,
                or_(Books.type == 'obs', Books.type == 'oper'),
                or_(Books.status == UNBOUND, Books.status == FAILED), 
            )
            if q.count() > 0:
                self.logger.info(
                    f"Not ready to register {book_id} due to unbound or "
                    "failed obs/oper books."
                )
                continue

            flist = self.get_files_for_stray_book(
                min_ctime= tc * 1e5,
                max_ctime= (tc + 1) * 1e5
            )
            if len(flist) > 0:
                stray_book = Books(
                    bid=book_id,
                    type="stray",
                    status=UNBOUND,
                    tel_tube=self.daq_node,
                    start=book_start,
                    stop=book_stop,
                    schema=0,
                )
                stray_book.path = self.get_book_path(stray_book)            
                self.logger.info(f"registering {book_id}")
                session.add(stray_book)
                session.commit()

    def get_book_abs_path(self, book):
        if book.path is None:
            book_path = self.get_book_path(book)
        else:
            book_path = book.path
        return os.path.join(self.output_root, book_path)

    def get_book_path(self, book):

        if book.type in ["obs", "oper"]:
            session_id = book.bid.split("_")[1]
            first5 = session_id[:5]
            odir = op.join(book.tel_tube, book.type, first5)
            return os.path.join(odir, book.bid)
        elif book.type in ["hk", "smurf"]:
            # get source directory for hk book
            first5 = book.bid.split("_")[1]
            assert first5.isdigit(), f"first5 of {book.bid} is not a digit"
            odir = op.join(book.tel_tube, book.type, book.bid)
            if book.type == 'smurf' and book.schema > 0:
                return odir + '.zip'
            return odir
        elif book.type in ["stray"]:
            first5 = book.bid.split("_")[1]
            assert first5.isdigit(), f"first5 of {book.bid} is not a digit"
            odir = op.join(book.tel_tube, book.type)
            return os.path.join(odir, book.bid)
        else:
            raise NotImplementedError(
                f"book type {book.type} not implemented"
            )

    def _get_binder_for_book(self, 
        book, 
        ignore_tags=False,
        ancil_drop_duplicates=False,
        allow_bad_timing=False,
        require_hwp=True,
        require_acu=True,
        require_monotonic_times=True,
        min_ctime=None, 
        max_ctime=None,
    ):
        """get the appropriate bookbinder for the book based on its type"""

        if book.type in ["obs", "oper"]:
            book_path = self.get_book_abs_path(book)

            # after sanity checks, now we proceed to bind the book.
            # get files associated with this book, in the form of
            # a dictionary of {stream_id: [file_paths]}
            filedb = self.get_files_for_book(book)

            # check if any of the files are too large. This is before
            # readout id checking because that one usually also raises
            # but for these we want to delete level 2 files
            if book.type == "obs":
                file_limit = MAX_OBS_LVL2_SIZE
            else:
                file_limit = MAX_OPER_LVL2_SIZE
            for _, flist in filedb.items():
                sz = np.array([os.path.getsize(f)/1e9 for f in flist])
                if np.any(sz > file_limit):
                    msg = "Files in book are too large:\n"
                    msg += "\n".join( [
                        f"{f},{round(s,2)} GB" for f,s in zip(flist, sz)
                    ])
                    raise FileTooLargeError(msg)

            obsdb = self.get_g3tsmurf_obs_for_book(book)
            readout_ids = self.get_readout_ids_for_book(book)
            hk_fields = self.config.get('hk_fields')

            if hk_fields is None:
                raise ValueError("`hk_fields` entry required for bookbinding.")

            # bind book using bookbinder library
            bookbinder = BookBinder(
                book, obsdb, filedb, self.lvl2_data_root, readout_ids, book_path, hk_fields,
                ignore_tags=ignore_tags,
                ancil_drop_duplicates=ancil_drop_duplicates,
                allow_bad_timing=allow_bad_timing,
                require_hwp=require_hwp,
                require_acu=require_acu,
                require_monotonic_times=require_monotonic_times,
                min_ctime=min_ctime, max_ctime=max_ctime,
            )
            return bookbinder

        elif book.type in ["hk", "smurf"]:
            # get source directory for hk book
            root = op.join(self.lvl2_data_root, book.type)
            timecode = book.bid.split("_")[1]
            assert timecode.isdigit(), f"timecode of {book.bid} is not a digit"
            book_path_src = op.join(root, timecode)

            # get target directory for hk book
            book_path_tgt = self.get_book_abs_path(book)
            odir, _ = op.split(book_path_tgt)
            if not op.exists(odir):
                os.makedirs(odir)
            
            bookbinder = TimeCodeBinder(
                book, timecode, book_path_src, book_path_tgt,
                ignore_pattern=SMURF_EXCLUDE_PATTERNS,
            )
            return bookbinder

        elif book.type in ["stray"]:
            flist = self.get_files_for_book(book)

            # get source directory for stray book
            root = op.join(self.lvl2_data_root, "timestreams")
            timecode = book.bid.split("_")[1]
            assert timecode.isdigit(), f"timecode of {book.bid} is not a digit"
            book_path_src = op.join(root, timecode)

           # get target directory for book
            book_path_tgt = self.get_book_abs_path(book)
            odir, _ = op.split(book_path_tgt)
            if not op.exists(odir):
                os.makedirs(odir)
                    
            bookbinder = TimeCodeBinder(
                book, timecode, book_path_src, book_path_tgt, 
                file_list=flist,
            )
            return bookbinder
        else:
            raise NotImplementedError(
                f"binder for book type {book.type} not implemented"
            )

    def _run_book_binding(
        self,
        book,
        session=None,
        message="",
        pbar=False,
        ignore_tags=False,
        ancil_drop_duplicates=False,
        allow_bad_timing=False,
        require_hwp=True,
        require_acu=True,
        require_monotonic_times=True,
        min_ctime=None, max_ctime=None,
        check_configs={}
    ):
        """Function that calls book-binding but does not update the database 
        this is implemented so be able to run things in parallel
        """
        if session is None:
            session = self.get_session()
        # get book id and book object, depending on whether book id is given or not
        if isinstance(book, Books):
            bid = book.bid
        elif isinstance(book, str):
            bid = book
            book = self.get_book(bid, session=session)
        else:
            raise NotImplementedError
        # check whether book exists in the database
        if not self.book_exists(bid, session=session):
            raise BookExistsError(f"Book {bid} does not exist in the database")
        # check whether book is already bound
        if (book.status == BOUND) :
            raise BookBoundError(f"Book {bid} is already bound")
        assert book.type in VALID_OBSTYPES

        err = None
        try:
            # find appropriate binder for the book type
            binder = self._get_binder_for_book(
                book, 
                ignore_tags=ignore_tags,
                ancil_drop_duplicates=ancil_drop_duplicates,
                allow_bad_timing=allow_bad_timing,
                require_acu=require_acu,
                require_hwp=require_hwp,
                require_monotonic_times=require_monotonic_times,
                min_ctime=min_ctime, max_ctime=max_ctime
            )
            binder.bind(pbar=pbar)
            
            # write M_index file
            if book.type in ['obs', 'oper']:
                tc = self.tube_configs[book.tel_tube]
            else:
                tc = {}
            binder.write_M_files(self.daq_node, tc)

            if book.type in ['obs', 'oper']:
                # check that detectors books were written out correctly
                self.logger.info("Checking Book {}".format(book.bid))
                check = BookScanner(
                    self.get_book_abs_path(book), config=check_configs
                )
                check.go()

            self.logger.info("Book {} bound".format(book.bid))
            status = BOUND
            
        except Exception as e:
            self.logger.error("Book {} failed".format(book.bid))
            err_msg = traceback.format_exc()
            self.logger.error(err_msg)
            message = f"{message}\ntrace={err_msg}" if message else err_msg
            status = FAILED
            err = e
        
        return book.bid, status, message, err

    def bind_book(
        self,
        book,
        session=None,
        message="",
        pbar=False,
        ignore_tags=False,
        ancil_drop_duplicates=False,
        allow_bad_timing=False,
        require_hwp=None,
        require_acu=True,
        require_monotonic_times=True,
        min_ctime=None, max_ctime=None,
        check_configs={}
    ):
        """Bind book using bookbinder

        Parameters
        ----------
        bid: str
            book id
        session: BookDB session
        message: string
            message to be added to the book
        test_mode : bool
            If in test_mode, this function will still run on already copied
            books, and will not update any db fields in the db. This is useful
            for testing purposes.
        ignore_tags : bool
            If true, book will be bound even if the tags between different level 2
            operations don't match. Only ever expected to be turned on by hand.
        ancil_drop_duplicates: if true, will drop duplicate data from ancilary  
            files. added to deal with an ocs aggregator error. Only ever 
            expected to be turned on by hand.
        allow_bad_timing: if true, will bind books even if the timing is low 
            precision
        require_hwp: bool, optional
            if True, requires that we find HWP data before binding the book. 
            hard-coded to False if self.daq_node is lat
        require_acu: bool, optional
            if True, requires that we have ACU data and that it has no dropouts 
            longer than 10s
        require_monotonic_times: bool, optional
            if True, requires that all HK data is monotonically increasing, 
            should never be set to False for obs books but less important for 
            oper books if there were ACU aggregation issues
        min_ctime: float, optional
            if not None, cuts the book down to have this minimum ctime
        max_ctime: float, optional
            if not None, cuts the book down to have this maximum ctime
        check_configs: dict
            additional non-default configurations to send to check book
        """
        if session is None:
            session = self.get_session()
        if require_hwp is None:
            require_hwp = self.require_hwp
        bid, status, message, err = self._run_book_binding(
            book,
            session=session,
            message=message,
            pbar=pbar,
            ignore_tags=ignore_tags,
            ancil_drop_duplicates=ancil_drop_duplicates,
            allow_bad_timing=allow_bad_timing,
            require_hwp=require_hwp,
            require_acu=require_acu,
            require_monotonic_times=require_monotonic_times,
            min_ctime=min_ctime, max_ctime=max_ctime,
            check_configs=check_configs
        )
        book = session.query(Books).filter(Books.bid == bid).one()
        book.status = status
        book.message = message
        session.commit()
        if status == FAILED:
            raise err


    def get_book(self, bid, session=None):
        """Get book from database.

        Parameters
        ----------
        bid: str
            book id
        session: BookDB session

        Returns
        -------
        book: book object or None if not found

        """
        if session is None:
            session = self.get_session()
        return session.query(Books).filter(Books.bid == bid).one_or_none()

    def get_books(self, session=None):
        """Get all books from database.

        Parameters
        ----------
        session: BookDB session

        Returns
        -------
        books: list of book objects

        """
        if session is None:
            session = self.get_session()
        return session.query(Books).all()

    def get_books_by_status(self, status, session=None):
        """Get all books with a given status from database.

        Parameters
        ----------
        session: BookDB session
        status: int
        
        Returns
        -------
        books: list of book objects

        """
        if session is None: session = self.get_session()
        return session.query(Books).filter(
            Books.status == status
        ).order_by(Books.start).all()

    # some aliases for readability
    def get_unbound_books(self, session=None):
        """Get all unbound books from database.

        Parameters
        ----------
        session: BookDB session

        Returns
        -------
        books: list of book objects

        """
        return self.get_books_by_status(UNBOUND, session)

    def get_bound_books(self, session=None):
        """Get all bound books from database.

        Parameters
        ----------
        session: BookDB session

        Returns
        -------
        books: list of book objects

        """
        return self.get_books_by_status(BOUND, session)
    
    def get_done_books(self, session=None):
        """Get all "done" books from database. Done means staged files are deleted.

        Parameters
        ----------
        session: BookDB session

        Returns
        -------
        books: list of book objects

        """
        return self.get_books_by_status(DONE, session)

    def get_failed_books(self, session=None):
        """Get all failed books from database

        Parameters
        ----------
        session: BookDB session

        Returns
        -------
        books: list of books

        """
        return self.get_books_by_status(FAILED, session)

    def get_rebind_books(self, session=None):
        """Get all books to be rebinded from database

        Parameters
        ----------
        session: BookDB session

        Returns
        -------
        books: list of books

        """
        return self.get_books_by_status(REBIND, session)

    def get_uploaded_books(self, session=None):
        """Get all books uploaded to librarian from database

        Parameters
        ----------
        session: BookDB session

        Returns
        -------
        books: list of books

        """
        return self.get_books_by_status(UPLOADED, session)

    def book_exists(self, bid, session=None):
        """Check if a book exists in the database.

        Parameters
        ----------
        session: BookDB session
        bid: str
            book id

        Returns
        -------
        boolean: True if book exists, False otherwise

        """
        if session is None:
            session = self.get_session()
        return self.get_book(bid, session=session) is not None

    def book_bound(self, bid, session=None):
        """Check if a book has been bound

        Parameters
        ----------
        bid: str
            book id
        session: BookDB session

        Returns
        -------
        boolean: True if book is bound, False otherwise

        """
        if session is None:
            session = self.get_session()
        book = self.get_book(bid, session=session)
        if book is None:
            return False
        return book.status == BOUND

    def commit(self, session=None):
        """Commit the book db.

        Parameters
        ----------
        session: BookDB session

        """
        if session is None:
            session = self.get_session()
        session.commit()

    def rollback(self, session=None):
        """Similar to commit, rollback the session

        Parameters
        ----------
        session: BookDB session

        """
        if session is None:
            session = self.get_session()
        session.rollback()

    @property
    def all_slots(self):
        return [x for xs in [
            t.get('slots') for (_,t) in self.tubes.items()
        ] for x in xs]

    def _find_incomplete(self, min_ctime, max_ctime, streams=None):
        """return G3tSmurf session query for incomplete observations
        """
        if streams is None:
            streams = self.all_slots

        session = self.get_g3tsmurf_session()
        q = session.query(G3tObservations).filter(
            G3tObservations.timestamp >= min_ctime,
            G3tObservations.timestamp <= max_ctime,
            G3tObservations.stream_id.in_(streams),
            or_(
                G3tObservations.stop == None,
                G3tObservations.stop >= dt.datetime.utcfromtimestamp(max_ctime),
            ),
        )
        return q

    @loop_over_tubes
    def update_bookdb_from_g3tsmurf(
        self,
        tube,
        min_ctime=None,
        max_ctime=None,
        min_overlap=30,
        ignore_singles=False,
        stream_ids=None,
        force_single_stream=False,
        return_obsset=False,
        incomplete_timeouts = (3,6),
    ):
        """Update bdb with new observations from g3tsmurf db.

        Parameters
        ----------
        tube: str
            which tube to look at. e.g., 'lati6', 'lati1', 'satp1'
        min_ctime: float
            minimum ctime to consider (in seconds since unix epoch)
        max_ctime: float
            maximum ctime to consider (in seconds since unix epoch)
        min_overlap: float
            minimum overlap between book and observation (in seconds)
        ignore_singles: boolean
            if True, ignore single observations
        stream_ids: list
            a list of stream ids (str) to consider. Defaults to list in
            imprinter configuration file.
        force_single_stream: boolean
            if True, treat observations from different streams separately
        return_obsset: boolean
            if True, return the list of observation sets instead of registering
            books. Useful as a debugging tool
        incomplete_timeouts: tuple
            hours for raising a (warning, error) if incomplete observations are
            found in the g3tsmurf database
        """
        if not self.build_det:
            return
        session, SMURF = self.get_g3tsmurf_session(return_archive=True)
        # set sensible ctime range is none is given
        if min_ctime is None:
            min_ctime = (
                session.query(G3tObservations.timestamp)
                .order_by(G3tObservations.timestamp)
                .first()[0]
            )
        if max_ctime is None:
            max_ctime = dt.datetime.now().timestamp()

        # get wafers
        if stream_ids is None:
            streams = self.tubes[tube].get("slots")
            if streams is None:
                raise ValueError(
                    f"Imprinter missing slot / stream_id" " information for {source}"
                )
        else:
            # if user input is present, format it like the query response
            streams = stream_ids
        self.logger.debug(f"Looking for observations from stream_ids {streams}")

        # check data transfer finalization
        final_time = SMURF.get_final_time(
            streams, min_ctime, max_ctime, check_control=True
        )
        if final_time < max_ctime:
            max_ctime = final_time
        self.logger.debug(f"Searching between {min_ctime} and {max_ctime}")

        # check for incomplete observations in time range
        q_incomplete = self._find_incomplete(min_ctime, max_ctime, streams)

        # if we have incomplete observations in our stream_id list we cannot
        # bookbind any observations overlapping the incomplete ones.
        if q_incomplete.count() > 0:
            new_ctime = min([obs.timestamp for obs in q_incomplete.all()])
            if max_ctime - new_ctime > 3600*incomplete_timeouts[1]:
                raise ValueError(
                    f"Found {q_incomplete.count()} incomplete observations. "
                    f"New max ctime would be {new_ctime}. More than "
                    f"{incomplete_timeouts[1]} hours in the past."
                )
            elif max_ctime - new_ctime > 3600*incomplete_timeouts[0]:
                level = self.logger.warning
            else:
                level = self.logger.debug
            level(
                f"Found {q_incomplete.count()} incomplete observations. "
                f"updating max ctime to {new_ctime}"
            )
            max_ctime = new_ctime

        min_start = dt.datetime.utcfromtimestamp(min_ctime)
        max_stop = dt.datetime.utcfromtimestamp(max_ctime)

        # find observations in time range that are already in books
        already_registered = [ x[0] for x in
            self.get_session().query(Observations.obs_id).join(Books).filter(
            Books.start >= min_start-dt.timedelta(hours=3),
            Books.stop <= max_stop+dt.timedelta(hours=3),
        ).all()]

        # find all complete observations that start within the time range and
        # are not already in books
        obs_q = session.query(G3tObservations).filter(
            G3tObservations.timestamp >= min_ctime,
            G3tObservations.timestamp < max_ctime,
            G3tObservations.stream_id.in_(streams),
            G3tObservations.stop < max_stop,
            not_(G3tObservations.stop == None),
            G3tObservations.obs_id.not_in(already_registered),
        )
        self.logger.debug(
            f"Found {obs_q.count()} level 2 observations to consider"
        )

        output = []
        def add_to_output(obs_list, mode):
            output.append(
                ObsSet( 
                    obs_list,
                    mode=mode,
                    slots=self.tubes[tube]["slots"],
                    tel_tube=tube,
                )
            )

        for stream in streams:
            # loop through all observations for this particular stream_id
            for str_obs in obs_q.filter(G3tObservations.stream_id == stream).all():
                # distinguish different types of books
                # operation book
                if get_obs_type(str_obs) == "oper":
                    add_to_output([str_obs], "oper")
                elif get_obs_type(str_obs) == "obs":
                    # force each observation to be its own book
                    if force_single_stream:
                        add_to_output([str_obs], "obs")
                    else:
                        # query for all possible types of overlapping 
                        # observations from other streams
                        q = obs_q.filter(
                            G3tObservations.stream_id != str_obs.stream_id,
                            G3tObservations.stream_id.in_(streams),
                            or_(
                                and_(
                                    G3tObservations.start <= str_obs.start,
                                    G3tObservations.stop >= str_obs.stop,
                                ),
                                and_(
                                    G3tObservations.stop >= str_obs.start,
                                    G3tObservations.stop <= str_obs.stop,
                                ),
                                and_(
                                    G3tObservations.start >= str_obs.start,
                                    G3tObservations.start <= str_obs.stop,
                                ),
                            ),
                        )

                        # if we failed to find any overlapping observations
                        if q.count() == 0 and not ignore_singles:
                            # append only this one when there's no overlapping 
                            # segments
                            add_to_output([str_obs], "obs")

                        elif q.count() > 0:
                            # obtain overlapping observations (returned as a tuple of stream_id)
                            obs_list = q.all()
                            # add the current obs too
                            obs_list.append(str_obs)

                            # remove overlapping operations
                            obs_list = [obs for obs in obs_list 
                                if get_obs_type(obs) == "obs"
                            ]
                            # check to make sure ALL observations overlap all 
                            # others and the overlap passes the minimum 
                            # requirement
                            overlap_time = (
                                np.min([o.stop for o in obs_list]) - 
                                np.max([o.start for o in obs_list])
                            )

                            if overlap_time.total_seconds() < 0:
                                continue
                            if overlap_time.total_seconds() < min_overlap:
                                continue

                            t_list = [o for o in obs_list if o.timing]
                            nt_list = [o for o in obs_list if not o.timing]
                            
                            if len(t_list)+len(nt_list) != len(obs_list):
                                raise ValueError(f"Cannot safely split timing"  
                                    f"info for {obs_list}"
                                )
                                

                            if len(t_list) > 0:
                                # add all of the possible overlaps
                                add_to_output(t_list, "obs")

                            if len(nt_list) > 0:
                                self.logger.debug(
                                    "registering single wafer books"       
                                    f" for {nt_list} because of low "
                                    "precision timing"
                                )
                                for obs in nt_list:
                                    add_to_output([obs], "obs")

        # remove exact duplicates in output
        output = drop_duplicates(output)
        self.logger.debug(f"Found {len(output)} possible books to register")
        # now our output is a list of ObsSet, where each ObsSet contains
        # all observations that overlap each other which should go
        # into the same book.
        if return_obsset:
            return output

        # look for repeat obs_ids
        obs_ids = []
        [obs_ids.extend( o.obs_ids) for o in output]
        if np.any([obs_ids.count(o) != 1 for o in obs_ids]):
            repeats = np.where( [obs_ids.count(o) != 1 for o in obs_ids])[0]
            raise OverlapObsError(f"Found repeated level 2 obs_ids in book "
                    f"list. {np.unique([obs_ids[x] for x in repeats])} overlap "
                    "multiple observations. Need to choose which books to put " 
                    "them in.")

        # register books in the book database (bdb)
        for oset in output:
            try:
                self.register_book(oset, commit=False)
            except BookExistsError:
                self.logger.debug(f"Book already registered: {oset}, skipping")
                continue
            except NoFilesError:
                self.logger.warning(f"No files found for {oset}, skipping")
                continue
        self.commit()

    def get_files_for_book(self, book):
        """Get all files in a book

        Parameters
        ----------
        book: Book object

        Returns
        -------
        files: dict if book.type is 'obs' or 'oper', otherwise list
            {obs_id: [file_paths...]}

        """
        if book.type in ["obs", "oper"]:
            session = self.get_g3tsmurf_session()
            obs_ids = [o.obs_id for o in book.obs]
            obs = (
                session.query(G3tObservations)
                .filter(G3tObservations.obs_id.in_(obs_ids))
                .all()
            )

            res = OrderedDict()
            for o in obs:
                res[o.obs_id] = sorted([f.name for f in o.files])
            return res
        elif book.type in ["stray"]:
            return self.get_files_for_stray_book(book)
        
        elif book.type == "hk":
            HK = self.get_g3thk()
            flist = (
                HK.session.query(HKFiles)
                .filter(
                    HKFiles.global_start_time >= book.start.timestamp(),
                    HKFiles.global_start_time < book.stop.timestamp(),
                )
                .all()
            )
            return [f.path for f in flist]
        elif book.type == "smurf":
            tcode = int(book.bid.split("_")[1])
            basepath = os.path.join(
                self.lvl2_data_root, 'smurf', str(tcode)
            )
            ignore = shutil.ignore_patterns(*SMURF_EXCLUDE_PATTERNS)
            flist = []
            for root, _, files in os.walk(basepath):
                to_ignore = ignore('', files)
                flist.extend([
                    os.path.join(basepath, root, f) 
                    for f in files if f not in to_ignore
                ])
            return flist

        else:
            raise NotImplementedError(
                f"book type {book.type} not understood for file search"
            )

    def get_files_for_stray_book(
        self, book=None, min_ctime=None, max_ctime=None
    ):
        """generate list of files that are not in detector books and should
        going into stray books. if book is None then we expect both min and max
        ctime to be provided

        Arguments
        ----------
        book: optional, book instance
        min_ctime: optional, minimum ctime value to search
        max_ctime: optional, maximum ctime value to search

        Returns
        --------
        list of files that should go into a stray book
        """
        if book is None:
            assert min_ctime is not None and max_ctime is not None
            start = dt.datetime.utcfromtimestamp(min_ctime)
            stop = dt.datetime.utcfromtimestamp(max_ctime)

            tcode = int(min_ctime//1e5)
            if max_ctime > (tcode+1)*1e5:
                self.logger.error(
                    f"Max ctime {max_ctime} is higher than would be expected "
                    f"for a single stray book with min ctime {min_ctime}. only"
                    " checking the first timecode directory"
                )
        else:
            assert book.type == 'stray'
            start = book.start
            stop = book.stop
            tcode = int(book.bid.split("_")[1])
        
        session = self.get_session()
        g3session, SMURF = self.get_g3tsmurf_session(return_archive=True)
        path = os.path.join(SMURF.archive_path, str(tcode))
        registered_obs = [ 
            x[0] for x in session.query(Observations.obs_id).join(Books).filter(
            Books.start >= start, 
            Books.start < stop,
            Books.status != WONT_BIND,
        ).all()]
        db_files = g3session.query(Files).filter(
            Files.name.like(f"{path}%")
        ).all()

        stray_files = []
        for f in db_files:
            if f.obs_id is None or f.obs_id not in registered_obs:
                stray_files.append(f.name)

        return stray_files
            

    def get_readout_ids_for_book(self, book):
        """
        Get all readout IDs for a book

        Parameters
        -----------
        book: Book object

        Returns
        -------
        readout_ids: dict
            {obs_id: [readout_ids...]}}

        """
        _, SMURF = self.get_g3tsmurf_session(return_archive=True)
        out = {}
        # load all obs and associated files
        for obs_id, files in self.get_files_for_book(book).items():
            self.logger.debug(f"Retrieving readout_ids for {obs_id}...")
            status = SmurfStatus.from_file(files[0])
            if not np.any(status.mask):
                raise MissingReadoutIDError(
                    f"Readout IDs not found for {obs_id}. SMuRF system was not "
                    "set up for science observations."
                )
            ch_info = get_channel_info(status, archive=SMURF)
            if "readout_id" not in ch_info:
                raise MissingReadoutIDError(
                    f"Readout IDs not found for {obs_id}. Indicates issue with G3tSmurf Indexing"
                )
            checks = ["NONE" in rid for rid in ch_info.readout_id]
            if np.all(checks):
                raise MissingReadoutIDError(
                    f"Readout IDs not found for {obs_id}. Indicates issue with G3tSmurf Indexing"
                )
            if np.any(checks):
                self.logger.warning(
                    f"Found {sum(checks)} channels without readout_id. Were fixed tones running?"
                )
            # make sure all rchannel ids are sorted
            assert list(ch_info.rchannel) == sorted(ch_info.rchannel)
            out[obs_id] = ch_info.readout_id
        return out

    def copy_smurf_files_to_book(self, book, book_path):
        """
        Copies smurf ancillary files to an operation book.

        Parameters
        -----------
        book : Books
            book object
        book_path : path
            Output path of book

        Returns
        ---------
        files : List[path]
            list of all metadata files copied to the book
        meta : dict[path]
            Dictionary of destination paths of important metadata such as
            tunes, bgmaps, or IVs.
        """
        if book.type != "oper":
            raise TypeError("Book must have type 'oper'")

        session, arc = self.get_g3tsmurf_session(
            book.tel_tube, return_archive=True
        )

        obs_ids = [o.obs_id for o in book.obs]
        obs = (
            session.query(G3tObservations)
            .filter(G3tObservations.obs_id.in_(obs_ids))
            .all()
        )

        files = []
        for ob in obs:
            files.extend(get_smurf_files(ob, arc.meta_path))

        dirname = os.path.join(book_path, "smurf")
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        meta_files = {}
        for f in files:
            basename = os.path.basename(f)
            dest = os.path.join(book_path, basename)
            shutil.copyfile(f, dest)

            if f.endswith("iv_analysis.npy"):
                meta_files["iv"] = basename
            elif f.endswith("bg_map.npy"):
                meta_files["bgmap"] = basename
            elif f.endswith("bias_step_analysis.npy"):
                meta_files["bias_steps"] = basename
            elif f.endswith("take_noise.npy"):
                meta_files["noise"] = basename
            elif f.endswith("bias_wave_analysis.npy"):
                meta_files["bias_wave"] = basename

        return files, meta_files

    def get_g3tsmurf_obs_for_book(self, book):
        """
        Get all g3tsmurf observations for a book

        Parameters
        -----------
        book: Book object

        Returns
        -------
        obs: dict
            {obs_id: G3tObservations}

        """
        session = self.get_g3tsmurf_session()
        obs_ids = [o.obs_id for o in book.obs]
        obs = (
            session.query(G3tObservations)
            .filter(G3tObservations.obs_id.in_(obs_ids))
            .all()
        )
        return {o.obs_id: o for o in obs}

    def _librarian_connect(self):
        """
        start connection to librarian
        """
        from hera_librarian import LibrarianClient
        from hera_librarian.settings import client_settings
        conn = client_settings.connections.get(
            self.config.get("librarian_conn")
        )
        if conn is None:
            raise ValueError(f"'librarian_conn' not in imprinter config")
        self.librarian = LibrarianClient.from_info(conn)

    def upload_book_to_librarian(self, book, session=None, raise_on_error=True):
        """Upload bound book to the librarian

        Parameters
        ----------
        book: Book object
        session: imprinter sqlalchemy session
            session that made book
        raise_on_error: bool
            raise an error if the librarian throws an error on the upload
        """

        if session is None:
            session = self.get_session()
        if self.librarian is None:
            self._librarian_connect()
        
        assert book.status == BOUND, "cannot upload unbound books"

        self.logger.info(f"Uploading book {book.bid} to librarian")
        try:     
            self.librarian.upload(
                Path( self.get_book_abs_path(book) ), 
                Path( book.path ), 
            )
            book.status = UPLOADED
            session.commit()
        except Exception as e:
            self.logger.error(
                f"Failed to upload book {book.bid}."
            )
            if raise_on_error:
                raise e
            else:
                return False, e
        return True, None
            
    def check_book_in_librarian(
        self, 
        book, 
        n_copies=1, 
        n_tries=1,
        raise_on_error=True
    ):
        """have the librarian validate the books is stored offsite. returns true
        if at least n_copies are storied offsite.
        """
        if self.librarian is None:
            self._librarian_connect()
        try:
            resp = self.librarian.validate_file(book.path)
            in_lib = sum(
                [(x.computed_same_checksum) for x in resp]
            ) >= n_copies
            if not in_lib:
                self.logger.info(f"received response from librarian {resp}")
                if n_tries > 1:
                    if book.type == 'smurf':
                        wait=30
                    else: 
                        wait=5
                    self.logger.warning(
                        f"Waiting {wait} seconds and trying book {book.bid} "
                        "with the librarian again"
                    )
                    time.sleep(wait)
                    return self.check_book_in_librarian(
                        book, n_copies=n_copies, 
                        n_tries=n_tries-1, 
                        raise_on_error=raise_on_error
                    )
        except Exception as e:
            if raise_on_error:
                raise e
            else: 
                self.logger.warning(
                    f"Failed to check libraian status for {book.bid}: {e}"
                )
                self.logger.warning(traceback.format_exc())
                in_lib = False
        return in_lib
        
    def delete_level2_files(self, book, verify_with_librarian=True,        
        n_copies_in_lib=2, n_tries=1, dry_run=True):
        """Delete level 2 data from already bound books

        Parameters
        ----------
        book: book object
        dry_run: bool
            if true, just prints plans to self.logger.info
        """
        if book.lvl2_deleted:
            self.logger.debug(
                f"Level 2 for {book.bid} has already been deleted"
            )
            return 0
        if book.status < UPLOADED:
            self.logger.warning(
                f"Book {book.bid} is not uploaded, not deleting level 2"
            )
            return 1
        if verify_with_librarian:
            in_lib = self.check_book_in_librarian(
                book, n_copies=n_copies_in_lib, n_tries=n_tries,
                raise_on_error=False
            )
            if not in_lib:
                self.logger.warning(
                    f"Book {book.bid} does not have {n_copies_in_lib} copies"
                    " will not delete level 2"
                )
                return 2
        
        self.logger.info(f"Removing level 2 files for {book.bid}")
        if book.type == "obs" or book.type == "oper":
            session, SMURF = self.get_g3tsmurf_session(
                return_archive=True
            )
            odic = self.get_g3tsmurf_obs_for_book(book)

            for oid, obs in odic.items():
                SMURF.delete_observation_files(
                    obs, session, dry_run=dry_run, my_logger=self.logger
                )
        elif book.type == "stray":
            session, SMURF = self.get_g3tsmurf_session(
                return_archive=True
            )
            flist = self.get_files_for_book(book)
            for f in flist:
                db_file = session.query(Files).filter(Files.name == f).one()
                SMURF.delete_file(
                    db_file, session, dry_run=dry_run, my_logger=self.logger
                )
        elif book.type == "smurf":
            tcode = int(book.bid.split("_")[1])
            basepath = os.path.join(
                self.lvl2_data_root, 'smurf', str(tcode)
            )
            if not dry_run:
                shutil.rmtree(basepath)

        elif book.type == "hk":
            HK = self.get_g3thk()
            flist = self.get_files_for_book(book)
            hkf_list = [
                HK.session.query(HKFiles).filter(
                    HKFiles.path == f
                ).one() for f in flist
            ]
            HK.batch_delete_files(
                hkf_list, dry_run=dry_run, my_logger=self.logger
            )
        else:
            raise NotImplementedError(
                f"Do not know how to delete level 2 files"
                f" for book of type {book.type}"
            )
        if not dry_run:
            book.lvl2_deleted = True
            self.session.commit()
        return 0

    def delete_book_staged(self, book, check_level2=False, 
        verify_with_librarian=False, n_copies_in_lib=1, override=False):
        """Delete all files associated with a book

        Parameters
        ----------
        book: Book object

        """
        if book.status == DONE:
            self.logger.debug(
                f"Book {book.bid} has already had staged files deleted"
            )
            return 0
        if not override:
            if book.status < UPLOADED:
                self.logger.warning(
                    "Cannot delete non-uploaded books without override"
                )
                return 1
        if check_level2 and not book.lvl2_deleted:
            self.logger.warning(
                f"Level 2 data not deleted for {book.bid}, not deleting "
                "staged"
            )
            return 2
        if verify_with_librarian:
            in_lib = self.check_book_in_librarian(
                book, n_copies=n_copies_in_lib, raise_on_error=False
            )
            if not in_lib:
                self.logger.warning(
                    f"Book {book.bid} does not have {n_copies_in_lib} copies"
                    " will not delete staged"
                )
                return 3

        # remove all files within the book
        book_path = self.get_book_abs_path(book)
        try:
            self.logger.info(
                f"Removing {book.bid} from staged"
            )
            if book.type == 'smurf' and book.schema == 1:
                os.remove(book_path)
            else:
                shutil.rmtree( book_path )
        except Exception as e:
            self.logger.warning(f"Failed to remove {book_path}: {e}")
            self.logger.error(traceback.format_exc())
            return 4
        book.status = DONE
        self.session.commit()
        return 0

    def find_missing_lvl2_obs_from_books(
        self, min_ctime, max_ctime
    ):
        """create a list of level 2 observation IDs that are not registered in 
        the imprinter database
        
        Arguments
        ----------
        min_ctime: minimum ctime value to search
        max_ctime: maximum ctime value to search

        Returns
        --------
        list of level 2 observation ids not in books
        """
        session = self.get_session()
        g3session, SMURF = self.get_g3tsmurf_session(return_archive=True)
        registered_obs = [
            x[0] for x in session.query(Observations.obs_id).join(Books).filter(
            Books.start >= dt.datetime.utcfromtimestamp(min_ctime), 
            Books.start < dt.datetime.utcfromtimestamp(max_ctime),
        ).all()]
        missing_obs = g3session.query(G3tObservations).filter(
            G3tObservations.timestamp >= min_ctime,
            G3tObservations.timestamp < max_ctime,
            G3tObservations.stream_id.in_(self.all_slots),
            G3tObservations.obs_id.not_in(registered_obs)
        ).all()
        return missing_obs

#####################
# Utility functions #
#####################

_primary_idx_map = {}


def get_frame_times(frame):
    """
    Returns timestamps for a G3Frame of detector data.

    Parameters
    --------------
    frame : G3Frame
        Scan frame containing detector data

    Returns
    --------------
    high_precision : bool
        If true, timestamps are computed from timing counters. If not, they are
        software timestamps

    timestamps : np.ndarray
        Array of timestamps for samples in the frame, in G3Time units (1e-8 sec)

    """
    if len(_primary_idx_map) == 0:
        for i, name in enumerate(frame["primary"].names):
            _primary_idx_map[name] = i

    c0 = frame["primary"].data[_primary_idx_map["Counter0"]]
    c2 = frame["primary"].data[_primary_idx_map["Counter2"]]

    if np.any(c0):
        return True, counters_to_timestamps(c0, c2) * core.G3Units.s
    else:
        return False, np.array(frame["data"].times)


def get_start_and_end(files):
    """
    Gets start and end time for a list of L2 detector data files

    Parameters
    -------------
    files : np.ndarray
        List of L2 G3 files for a single detector set in an observation

    Returns
    --------
    t0 : float
        Start time of observation
    t1 : float
        End time or observation
    """
    _files = sorted(files)

    # Get start time
    found_scan = False
    for file in _files:
        for frame in core.G3File(file):
            if frame.type == core.G3FrameType.Scan:
                t0 = get_frame_times(frame)[1][0]
                found_scan = True
                break
        if found_scan:
            break

    # Get end time
    frame = None
    found_scan = False
    for file in _files[::-1]:
        for _frame in core.G3File(file):
            if _frame.type == core.G3FrameType.Scan:
                frame = _frame
                found_scan = True

        if found_scan:
            break

    t1 = get_frame_times(frame)[1][-1]

    return t0, t1


def create_g3tsmurf_session(config):
    """create a connection session to g3tsmurf database

    Parameters
    ----------
    config: str
        path to a yaml file that contains g3tsmurf db conn details
        or it could be a dictionary with keys, etc.

    Returns
    -------
    session: sqlalchemy session
        a session to the g3tsmurf database

    """
    # create database connection
    config = load_configs(config)
    config["db_args"] = {"connect_args": {"check_same_thread": False}}
    SMURF = G3tSmurf.from_configs(config)
    session = SMURF.Session()
    return session, SMURF


def stream_timestamp(obs_id):
    """parse observation id to obtain a (stream, timestamp) pair

    Parameters
    ----------
    obs_id: str
        observation id

    Returns
    -------
    stream: str
        stream id
    timestamp: str
        timestamp of the observation

    """
    return "_".join(obs_id.split("_")[1:-1]), obs_id.split("_")[-1]


def get_obs_type(obs: G3tObservations):
    """Get the type of observation based on the observation id

    Parameters
    ----------
    obs: G3tObservations object

    Returns
    -------
    obs_type: str

    """
    if obs.tag is None:
        return None
    tags = obs.tag.split(",")
    # assert tags[0] in VALID_OBSTYPES, f"Unknown tag {tags[0]}"  # this is too strict for now
    return tags[0]


class ObsSet(list):
    """A thin wrapper around a list of observations that are all
    part of the same book"""

    def __init__(self, *args, mode="obs", slots=None, tel_tube=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.mode = mode
        self.slots = slots  # all possible slots (not those in the ObsSet)
        self.tel_tube = tel_tube

    @property
    def obs_ids(self):
        """list of observation ids"""
        return sorted([obs.obs_id for obs in self])

    def contains_stream(self, stream_id):
        """check whether an ObsSet has a given stream id

        Parameters
        ----------
        obsset: dict with obs ids as keys

        Returns
        -------
        boolean

        """
        for obs_id in self.obs_ids:
            if stream_id in stream_timestamp(obs_id)[0]:
                return True
        return False

    def __eq__(self, other):
        """check if two ObsSets are equal"""
        return self.obs_ids == other.obs_ids

    def get_id(self, mode=None):
        """get a unique id for this ObsSet. Following the specification
        in the TOD format document.

        Parameters
        ----------
        mode: str
            observation mode / type, e.g., obs, oper, etc.

        Returns
        -------
        id: str
            unique id for this ObsSet, used to identify the book in the book database

        """
        # if no mode is specified, use the one provided for the set
        if mode is None:
            mode = self.mode

        assert mode in VALID_OBSTYPES, f"Invalid mode {mode}"

        # common values across modes
        # parse observation time and name the book with the first timestamp
        _, timestamp = stream_timestamp(self.obs_ids[0])

        # the rest of the id depends on the mode
        if mode in ["obs", "oper"]:
            # get slot flags
            slot_flags = ""
            for slot in self.slots:
                slot_flags += "1" if self.contains_stream(slot) else "0"
            bid = f"{mode}_{timestamp}_{self.tel_tube}_{slot_flags}"
        else:
            raise NotImplementedError

        return bid

    def __str__(self):
        return f"{self.get_id()}: {self.obs_ids}"


def drop_duplicates(obsset_list: List[ObsSet]):
    """Drop duplicate obssets from a list of obssets

    Parameters
    ----------
    obsset_list: list of ObsSet objects

    Returns
    -------
    obsset_list: list of ObsSet objects
        with duplicates removed
    """
    ids_list = [oset.obs_ids for oset in obsset_list]
    new_obsset_list = list(
        OrderedDict(
            (tuple(ids), oset) for (ids, oset) in zip(ids_list, obsset_list)
        ).values()
    )
    return new_obsset_list


def require(dct, keys):
    """
    Check whether a dictionary contains all the required keys.

    Parameters
    ----------
    dct: dict
        dictionary to check
    keys: list
        list of required keys
    """
    if dict is None:
        raise ValueError("Missing required dictionary")
    for k in keys:
        if k not in dct:
            raise ValueError(f"Missing required key {k}")



import unittest
import numpy as np

from sotodlib import core
from sotodlib.hwp.hwp import get_tau_hwp


def get_hwp_spin_up_down(tod, spin_up, sign):
    """ Make spin up down hwp_angle """
    if spin_up:
        f0, f1 = 0, 2
    else:
        f0, f1 = 2, 0
    hwp_rate = np.linspace(f0, f1, tod.samps.count)
    hwp_angle = np.cumsum(np.ones(tod.samps.count) / 200
                          * sign * hwp_rate * 2 * np.pi)
    tod.wrap('hwp_angle', np.mod(hwp_angle, 2 * np.pi),
             axis_map=[(0, 'samps')])


def make_hwp_spin_up_down_tod(
    taus=np.array([1e-3, 2e-3]),
    A=3e-2,
    wn=5e-5,
    spin_up=True,
    sign=1,
):
    """
    Make an axis manager of hwp spin up down observation
    """
    t = np.arange(200 * 600)/200.  # 10 mins

    dets = [f'det{i}' for i in range(len(taus))]

    tod = core.AxisManager(core.LabelAxis('dets', vals=dets),
                           core.OffsetAxis('samps', count=len(t)))
    tod.wrap('timestamps', t, axis_map=[(0, 'samps')])
    get_hwp_spin_up_down(tod, spin_up, sign)

    hwp_rate = np.gradient(np.unwrap(tod.hwp_angle)) * 200
    signal = A * np.cos(4 * (tod.hwp_angle - hwp_rate * taus[:, None]))
    signal = np.float32(signal)
    np.random.seed(0)
    signal += np.random.normal(0, wn * np.sqrt(200),
                               (tod.dets.count, tod.samps.count))
    tod.wrap('signal', signal, axis_map=[(0, 'dets'), (1, 'samps')])
    return tod


class TauHWPTest(unittest.TestCase):
    """ Test the tau_hwp fitting functions """
    def test_fit(self):
        taus = np.array([1e-3, 2e-3])

        tod = make_hwp_spin_up_down_tod(taus=taus, spin_up=True, sign=1)
        result = get_tau_hwp(tod, full_output=True)
        self.assertAlmostEqual(result.tau_hwp[0], taus[0], delta=1e-4)
        self.assertAlmostEqual(result.tau_hwp[1], taus[1], delta=1e-4)

        tod = make_hwp_spin_up_down_tod(taus=taus, spin_up=False, sign=1)
        result = get_tau_hwp(tod)
        self.assertAlmostEqual(result.tau_hwp[0], taus[0], delta=1e-4)
        self.assertAlmostEqual(result.tau_hwp[1], taus[1], delta=1e-4)

        tod = make_hwp_spin_up_down_tod(taus=taus, spin_up=True, sign=-1)
        result = get_tau_hwp(tod)
        self.assertAlmostEqual(result.tau_hwp[0], taus[0], delta=1e-4)
        self.assertAlmostEqual(result.tau_hwp[1], taus[1], delta=1e-4)

        tod = make_hwp_spin_up_down_tod(taus=taus, spin_up=False, sign=-1)
        result = get_tau_hwp(tod)
        self.assertAlmostEqual(result.tau_hwp[0], taus[0], delta=1e-4)
        self.assertAlmostEqual(result.tau_hwp[1], taus[1], delta=1e-4)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simulate the nominal hardware model.
"""

import sys
import argparse

from collections import OrderedDict

from ..sim_hardware import sim_nominal, sim_detectors_toast


def main():
    parser = argparse.ArgumentParser(
        description="This program simulates the current nominal hardware\
            model and dumps it to disk",
        usage="so_hardware_sim [options] (use --help for details)")

    parser.add_argument(
        "--out", required=False, default="hardware",
        help="Name (without extensions) of the output hardware file"
    )

    parser.add_argument(
        "--plain", required=False, default=False, action="store_true",
        help="Write plain text (without gzip compression)"
    )

    parser.add_argument(
        "--overwrite", required=False, default=False, action="store_true",
        help="Overwrite any existing output file."
    )

    args = parser.parse_args()

    print("Getting nominal config...", flush=True)
    hw = sim_nominal()
    for tele, teleprops in hw.data["telescopes"].items():
        print("Simulating detectors for telescope {}...".format(tele),
              flush=True)
        sim_detectors_toast(hw, tele)

    if args.plain:
        outpath = "{}.toml".format(args.out)
        print("Dumping config to {}...".format(outpath))
        hw.dump(outpath, overwrite=args.overwrite, compress=False)
    else:
        outpath = "{}.toml.gz".format(args.out)
        print("Dumping config to {}...".format(outpath))
        hw.dump(outpath, overwrite=args.overwrite, compress=True)

    return



# Copyright (c) 2024-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""High-level blocks of workflow tools.
"""
import toast
from toast.observation import default_values as defaults

from .. import workflows as wrk


def setup_load_or_simulate_observing(parser, operators):
    # Loading data from disk
    wrk.setup_load_data_hdf5(operators)
    wrk.setup_load_data_books(operators)
    wrk.setup_load_data_context(operators)
    wrk.setup_act_responsivity_sign(operators)
    wrk.setup_az_intervals(operators)
    wrk.setup_weather_model(operators)
    wrk.setup_diff_noise_estimation(operators)
    wrk.setup_readout_filter(operators)
    wrk.setup_deconvolve_detector_timeconstant(operators)

    # Simulated observing
    wrk.setup_simulate_observing(parser, operators)
    wrk.setup_simple_noise_models(operators)

    wrk.setup_pointing(operators)


def load_or_simulate_observing(job, otherargs, runargs, comm):
    log = toast.utils.Logger.get()
    job_ops = job.operators

    # Timer for reporting the progress
    timer = toast.timing.Timer()
    timer.start()

    if job_ops.sim_ground.enabled or otherargs.schedule is not None:
        data = wrk.simulate_observing(job, otherargs, runargs, comm)
        wrk.select_pointing(job, otherargs, runargs, data)
        wrk.simple_noise_models(job, otherargs, runargs, data)
    else:
        group_size = wrk.reduction_group_size(job, runargs, comm)
        toast_comm = toast.Comm(world=comm, groupsize=group_size)
        data = toast.Data(comm=toast_comm)
        # Load data from all formats
        wrk.load_data_hdf5(job, otherargs, runargs, data)
        wrk.load_data_books(job, otherargs, runargs, data)
        wrk.load_data_context(job, otherargs, runargs, data)
        wrk.act_responsivity_sign(job, otherargs, runargs, data)
        wrk.create_az_intervals(job, otherargs, runargs, data)
        wrk.apply_readout_filter(job, otherargs, runargs, data)
        wrk.deconvolve_detector_timeconstant(job, otherargs, runargs, data)
        # Append a weather model
        wrk.append_weather_model(job, otherargs, runargs, data)
        # Before running simulations based on real data, we need
        # a starting noise estimate
        wrk.diff_noise_estimation(job, otherargs, runargs, data)
        # optionally zero out
        if hasattr(otherargs, "zero_loaded_data") and otherargs.zero_loaded_data:
            toast.ops.Reset(detdata=[defaults.det_data]).apply(data)
        wrk.select_pointing(job, otherargs, runargs, data)

    job_ops.mem_count.prefix = "After Data Load or Simulate"
    job_ops.mem_count.apply(data)

    return data


def setup_preprocess(parser, operators):
    wrk.setup_filter_hwpss(operators)
    wrk.setup_filter_hwpss_model(operators)
    wrk.setup_simple_jumpcorrect(operators)
    wrk.setup_simple_deglitch(operators)
    wrk.setup_flag_diff_noise_outliers(operators)
    wrk.setup_flag_noise_outliers(operators)


def preprocess(job, otherargs, runargs, data):
    """Apply flags and cuts based on data quality."""
    log = toast.utils.Logger.get()
    job_ops = job.operators

    # Run these on the original data
    wrk.simple_jumpcorrect(job, otherargs, runargs, data)
    wrk.simple_deglitch(job, otherargs, runargs, data)

    if otherargs.preprocess_copy:
        prename = "preprocess"
        toast.ops.Copy(detdata=[(defaults.det_data, prename)]).apply(data)
    else:
        prename = defaults.det_data

    def _run_pre_op(op, wrkf):
        if not op.enabled:
            return
        save_op_detdata = op.det_data
        op.det_data = prename
        wrkf(job, otherargs, runargs, data)
        op.det_data = save_op_detdata

    # Filters which we might want to run on a copy of the data.
    _run_pre_op(job_ops.hwpfilter, wrk.filter_hwpss)
    _run_pre_op(job_ops.hwpss_model, wrk.filter_hwpss_model)
    _run_pre_op(job_ops.diff_noise_cut, wrk.flag_diff_noise_outliers)
    _run_pre_op(job_ops.noise_cut, wrk.flag_noise_outliers)

    if otherargs.preprocess_copy:
        toast.ops.Delete(detdata=[prename,]).apply(data)

    # If we generated a relative gain from the HWPSS 2f component,
    # apply that now.
    if job_ops.hwpss_model.enabled:
        if job_ops.hwpss_model.relcal_fixed is not None:
            toast.ops.CalibrateDetectors(
                cal_name=job_ops.hwpss_model.relcal_fixed
            ).apply(data)
        elif job_ops.hwpss_model.relcal_continuous is not None:
            toast.ops.Combine(
                first=defaults.det_data,
                second=job_ops.hwpss_model.relcal_continuous,
                result=defaults.det_data,
                op="multiply",
            ).apply(data)




"""
Functions for simulating glitches, jumps and other 
bad features in the data
"""
import numpy as np
from so3g.proj import Ranges, RangesMatrix


def add_random_glitches(tod, params={}, signal='glitches', flag='true_glitches',
                        overwrite=False, verbose=False):
    """Add glitches (spikes that just effect one data point) to the data.
    
    Args:
        tod (AxisManager): the tod 
        params (dictionary): Use to overwrite the default values
                n_glitch: the expected number of glitches per detector per observation
                sig_n_glitch: the width of the expected distribution
                h_glitch: the expected higher of the glitches
                sig_h_glitch: the expected higher of the glitches
        signal (string): name of the place to add the badness to the tod
        flag (string): name of the flag to store where the glitches are
        overwrite (bool): if true, write over signal. if false, add to signal
        verbose (bool): print the number of glitches added
    """
    gparams = {'n_glitch' : 1, 'sig_n_glitch' : 0.1, 
               'h_glitch' : 10, 'sig_h_glitch' : 0.01}
    gparams.update(params)
    params=gparams
    
    n_tot = int(np.abs(params['sig_n_glitch']*tod.dets.count*np.random.randn(1) 
                       + params['n_glitch']*tod.dets.count))
    places = np.random.randint( tod.dets.count*tod.samps.count, size=(n_tot,))
    heights = np.random.randn( n_tot)*params['sig_h_glitch'] + params['h_glitch']
    glitches = np.zeros( (tod.dets.count*tod.samps.count,) )
    glitches[places] = heights
    
    glitches = np.reshape( glitches, (tod.dets.count, tod.samps.count) )
    truth = RangesMatrix( [Ranges.from_mask( g!=0) for g in glitches])
    
    if verbose:
        print('Adding {} glitches to {} detectors'.format(n_tot, 
                                                          np.sum( [len(t.ranges())>0 for t in truth]),))
                                                                      
    if not signal in tod:
        tod.wrap(signal, glitches,
                 [(0,tod.dets), (1,(tod.samps))] )
    else:
        if overwrite:
            tod[signal] = glitches
        else:
            tod[signal] += glitches
            
    if not flag in tod.flags:
        tod.flags.wrap(flag, truth)
    else:
        if overwrite:
            tod.flags[flag] = truth
        else:
            tod.flags[flag] += truth    

def add_random_jumps(tod, params={}, signal='jumps', flag='true_jumps',
                        overwrite=False, verbose=False):
    """Add jumps (changes in DC level) to the data.
    
    Args:
        tod (AxisManager): the tod 
        params (dictionary): Use to overwrite the default values
                n_jump is the expected number of jumps per detector per observation
               sig_n_jump is the width of the expected distribution
               h_jump is the expected higher of the jumps
               sig_h_jump is the expected higher of the jumps
        signal (string): name of the place to add the badness to the tod
        flag (string): name of the flag to store where the glitches are
        overwrite (bool): if true, write over signal. if false, add to signal
        verbose (bool): print the number of glitches added
    """
  
    gparams = {'n_jump' : 0.2, 'sig_n_jump' : 0.05, 
               'h_jump' : 0.1, 'sig_h_jump' : 0.3}
    gparams.update(params)
    params=gparams
    
    n_tot = int(np.abs(params['sig_n_jump']*tod.dets.count*np.random.randn(1) 
                       + params['n_jump']*tod.dets.count))
    places = np.random.randint( tod.dets.count*tod.samps.count, size=(n_tot,))
    heights = np.random.randn( n_tot)*params['sig_h_jump'] + params['h_jump']
    jumps = np.zeros( (tod.dets.count*tod.samps.count,) )
    jumps[places] = heights
    
    jumps = np.reshape( jumps, (tod.dets.count, tod.samps.count) )
    truth = RangesMatrix( [Ranges.from_mask( g!=0) for g in jumps])
    ## actually make these jumps and not glitches
    jumps = np.cumsum(jumps, axis=1)    
    
    if verbose:
        print('Adding {} jumps to {} detectors'.format(n_tot, 
                                                          np.sum( [len(t.ranges())>0 for t in truth]),))
                                                                      
    if not signal in tod:
        tod.wrap(signal, jumps,
                 [(0,tod.dets), (1,(tod.samps))] )
    else:
        if overwrite:
            tod[signal] = jumps
        else:
            tod[signal] += jumps
            
    if not flag in tod.flags:
        tod.flags.wrap(flag, truth)
    else:
        if overwrite:
            tod.flags[flag] = truth
        else:
            tod.flags[flag] += truth

def add_random_offsets( tod, params={}, signal='offset',
                        overwrite=False, verbose=False):
    """Add uniformly distributed DC offsets to the data
    
    Args:
        tod (AxisManager): the tod 
        params (dictionary): Use to overwrite the default values
                offset_low: lowest possible offset
                offset_high: highest possible offset
        signal (string): name of the place to add the badness to the tod
        overwrite (bool): if true, write over signal. if false, add to signal
        verbose (bool): print that offsets are added
    """
    
    gparams = {'offset_low' : -20, 
               'offset_high':  20 }
    gparams.update(params)
    params=gparams
    
    offsets = params['offset_low'] + np.random.rand(tod.dets.count)*(params['offset_high']
                                                                     -params['offset_low'])
    offs = np.zeros( (tod.dets.count, tod.samps.count))
    offs += offsets[:,None]
    
    if verbose:
        print('Adding Offsets to detectors')
                                                                      
    if not signal in tod:
        tod.wrap(signal, offs,
                 [(0,tod.dets), (1,(tod.samps))] )
    else:
        if overwrite:
            tod[signal] = offs
        else:
            tod[signal] += offs

def add_random_trends( tod, params={}, signal='trends',
                        overwrite=False, verbose=False):
    """Add normally distributed trends offsets to the data
    
    Args:
        tod (AxisManager): the tod 
        params (dictionary): Use to overwrite the default values
                slope_mean: normal distribution mean
                slope_sig: normal distribution offset
        signal (string): name of the place to add the badness to the tod
        overwrite (bool): if true, write over signal. if false, add to signal
        verbose (bool): print that offsets are added
    """
    gparams = {'slope_mean' : 0.1, 
               'slope_sig':  0.01 }
    gparams.update(params)
    params=gparams
    
    s = np.random.randn(tod.dets.count)*params['slope_sig']+params['slope_mean']
    
    slopes = np.zeros( (tod.dets.count, tod.samps.count))
    slopes += s[:,None]*np.linspace(0,1,tod.samps.count)[None,:]
    
    if verbose:
        print('Adding Trends to Detectors')
                                                                      
    if not signal in tod:
        tod.wrap(signal, slopes,
                 [(0,tod.dets), (1,(tod.samps))] )
    else:
        if overwrite:
            tod[signal] = slopes
        else:
            tod[signal] +=slopes



# Copyright (c) 2018-2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Data simulation.

This module contains code for simulating data.

"""

import numpy as np

from spt3g import core
from .core.g3_core import DataG3Module


class PipelineSeeder(list):
    """
    A way to introduce statically generated frames into a pipeline.
    Instantiate this as a list of seed Frames, then add it as the
    first Pipeline element.
    """
    def __call__(self, frame_in):
        output = []
        if frame_in is not None:
            output.append(frame_in)
        if len(self):
            output.append(self.pop(0))
        return output

def enumerate_det_id(n_dets, freq=39, band='LF1'):
    """
    Generator for looping through detector names. Not very smart, only for
    basic testing.

    Args:
        n_dets (int): number of detector names to make
        freq (int): detector freqency
        band (str): detector band

    Returns:
        SO formatted detector keys
    """
    types = ['A', 'B']
    for n in range(n_dets):
        dnum = int(np.floor(n/2))
        t = int(np.mod(n,2))
        yield '{}_{:03}_{}_{}'.format(freq,dnum,band,types[t])

def noise_scan_frames(n_frames=3, n_dets=20, input='signal', n_samps=200,
                      samp_rate=0.005*core.G3Units.second,
                      t_start=core.G3Time('2020-1-1T00:00:00')):
    """
    Generate a list of frames filled with noise data and nothing else.

    Args:
        n_frames (int): number of frames to make
        n_dets (int): number of detectors per frame
        input (str): name of G3TimestreamMap for detectors, should be some form of 'signal'
        n_samps (int): number of samples per detector timestream
        samp_rate (G3Unit.second): detector sampling rate
        t_start (G3Time): start time of the set of frames
    """
    frame_list = []
    for n in range(n_frames):
        f = core.G3Frame()
        f.type = core.G3FrameType.Scan
        tsm = core.G3TimestreamMap()
        z = np.zeros( (n_samps,) )
        for d in enumerate_det_id(n_dets):
            tsm[d] = core.G3Timestream(z)
        tsm.start = t_start
        tsm.stop = t_start + n_samps*samp_rate
        tsm = MakeNoiseData().apply(tsm)
        f[input] = tsm
        t_start += n_samps*samp_rate
        frame_list.append(f)
    return frame_list

class MakeNoiseData(DataG3Module):
    """
    Writes a signal with just noise. To be used where an observation
    has already been set up but there's no data (such as the output of
    so3g.python.quicksim.py) mostly just an easy way to get numbers in G3Timestreams

    The noise is a basic white noise plus a 1/f component described by a
    knee frequency and index

    Args:
        input (str): the key to a G3Timestream map in the G3Frame
            to replace with noise data
        output (str or None): key of G3Timestream map of output data
            if None: input will be overwritten with output
        white_noise (float): while noise level
        f_knee (float): knee frequency
        f_knee_index (float): index of 1/f spectrum, should be negative

    Returns:
        None
    """
    def __init__(self, input='signal', output=None, white_noise = 0.005,
                 f_knee = 0.01, f_knee_index=-2):
        self.white_noise = white_noise
        self.f_knee = f_knee
        self.f_knee_index = f_knee_index
        super().__init__(input, output)

    def process(self, data, k):
        freqs = np.fft.fftfreq(data.n_samples, core.G3Units.Hz/data.sample_rate)
        noise = self.white_noise*(1 + (freqs[1:]/self.f_knee)**self.f_knee_index)
        noise = noise*np.exp( 1.0j * np.random.uniform(0, 2*np.pi, size=(data.n_samples-1),))
        ## prevent divide by zero error and assume 1/f doesn't go to infinity
        noise = np.append(noise[0], noise)
        return np.real(np.fft.fft(noise)).astype('float64')

class MakeJumps(DataG3Module):

    """
    G3Module that takes a G3Timestream map and adds randomly
    distributed jumps

    Args:
        input (str): the key to a G3TimestreamMap that is the data source
        output (str or None): the key for a G3TimestreamMap that will have data
            plus glitches. If None: jumps are added to Input
        info (str): a G3Timestream map will be made with this name that
            includes just the jumps.
        max_jumps (int): number of jumps in each G3Timestream is
            np.random.randint(max_jumps)
        height_std_sigma (float): hight of each jump is a draw from a normal
            distribution with standard deviation of
            height_std_sigma*np.std(timestream)
    """

    def __init__(self, input='signal', output=None, info='flags_encoded_jumps',
                 max_jumps=3, height_std_sigma=10):
        self.info = info
        self.max_jumps = max_jumps
        self.height_std_sigma = height_std_sigma
        super().__init__(input, output)

    def __call__(self, f):
        if f.type == core.G3FrameType.Scan:
            self.jump_map = core.G3TimestreamMap()

        super().__call__(f)

        if f.type == core.G3FrameType.Scan:
            self.jump_map.start = f[self.input].start
            self.jump_map.stop = f[self.input].stop
            f[self.info] = self.jump_map

    def process(self, data, det_name):
        locs = np.random.randint(data.n_samples, size=(np.random.randint(self.max_jumps),) )
        heights = np.random.randn( len(locs) )*self.height_std_sigma*np.std(data)
        jumps = np.zeros( (data.n_samples,) )
        for i in range(len(locs)):
            jumps[locs[i]:] += heights[i]

        self.jump_map[det_name] = core.G3Timestream( jumps )
        return data + jumps


class MakeGlitches(DataG3Module):
    """
    G3Module that takes the G3Timestream map and adds randomly
    distributed glitches

    Args:
        input (str): the key to a G3TimestreamMap that is the data source
        output (str or None): the key for a G3TimestreamMap that will have data
            plus glitches. If None: Glitches are added to Input
        info (str): a G3Timestream map will be made with this name that
            includes just the glitches.
        max_glitches (int): number of glitches in each G3Timestream is
            np.random.randint(max_glitches)
        height_std_sigma (float): hight of each jump is a draw from a normal
            distribution with standard deviation of
            height_std_sigma*np.std(timestream)
    """

    def __init__(self, input='signal', output=None, info='flags_encoded_glitches', max_glitches=3, height_std_sigma=20):
        self.info = info
        self.max_glitches = max_glitches
        self.height_std_sigma = height_std_sigma
        super().__init__(input, output)

    def __call__(self, f):
        if f.type == core.G3FrameType.Scan:
            self.glitch_map = core.G3TimestreamMap()

        super().__call__(f)

        if f.type == core.G3FrameType.Scan:
            self.glitch_map.start = f[self.input].start
            self.glitch_map.stop = f[self.input].stop
            f[self.info] = self.glitch_map

    def process(self, data, det_name):
        locs = np.random.randint(data.n_samples, size=(np.random.randint(self.max_glitches),) )
        heights = np.random.randn( len(locs) )*self.height_std_sigma*np.std(data)
        glitches = np.zeros( (data.n_samples,) )
        glitches[locs] += heights

        self.glitch_map[det_name] = core.G3Timestream( glitches )
        return core.G3Timestream( data + glitches )



#!/usr/bin/env python3

# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""
This workflow converts data between supported TOAST data formats.  It uses a single
process group to load data from:

- A set of HDF5 observation files

- A set of books

And then writes these out to:

- A set of HDF5 observation files

- A set of books

You can see the automatically generated command line options with:

    toast_so_convert --help

Or you can dump a config file with all the default values with:

    toast_so_convert --default_toml config.toml

This script contains just comments about what is going on.  For details about all the
options for a specific Operator, see the documentation or use the help() function from
an interactive python session.

"""

import argparse
import datetime
import os

import numpy as np

from astropy import units as u

# Import sotodlib.toast first, since that sets default object names
# to use in toast.
import sotodlib.toast as sotoast

import toast
import toast.ops
from toast.mpi import MPI, Comm
from toast.observation import default_values as defaults

from .. import ops as so_ops
from .. import workflows as wrk


def main():
    env = toast.utils.Environment.get()
    log = toast.utils.Logger.get()
    gt = toast.timing.GlobalTimers.get()
    gt.start("toast_so_convert (total)")
    timer = toast.timing.Timer()
    timer.start()

    # Get optional MPI parameters
    comm, procs, rank = toast.get_world()

    # This workflow is I/O dominated and does not benefit from
    # threads.  If the user has not told us to use multiple threads,
    # then just use one.

    if "OMP_NUM_THREADS" in os.environ:
        nthread = os.environ["OMP_NUM_THREADS"]
    else:
        nthread = 1
    log.info_rank(
        f"Executing workflow with {procs} MPI tasks, each with "
        f"{nthread} OpenMP threads at {datetime.datetime.now()}",
        comm,
    )

    mem = toast.utils.memreport(msg="(whole node)", comm=comm, silent=True)
    log.debug_rank(f"Start of the workflow:  {mem}", comm)

    # Argument parsing and operator configuration

    parser = argparse.ArgumentParser(description="Convert SO data between formats")
    parser.add_argument(
        "--timing_root",
        required=False,
        default=None,
        help="Root filename (without extension) for logging timing info",
    )
    parser.add_argument(
        "--config_log",
        required=False,
        default=None,
        help="Write final config file to this path",
    )

    operators = list()
    wrk.setup_load_data_books(operators)
    wrk.setup_load_data_hdf5(operators)
    wrk.setup_load_data_context(operators)
    wrk.setup_save_data_books(operators)
    wrk.setup_save_data_hdf5(operators)

    job, config, otherargs, runargs = wrk.setup_job(parser=parser, operators=operators)

    # We enforce only one process group for this job
    group_size = procs
    if runargs.group_size is not None and runargs.group_size != group_size:
        raise RuntimeError(
            "This conversion script only works with a single process group"
        )

    # Create the toast communicator
    toast_comm = toast.Comm(world=comm, groupsize=group_size)

    # Log the config that was actually used at runtime.
    if otherargs.config_log is not None:
        toast.config.dump_toml(otherargs.config_log, config, comm=comm)

    # If this is a dry run, exit
    if otherargs.dry_run:
        log.info_rank("Dry-run complete", comm=comm)
        return

    # Load data
    data = toast.Data(comm=toast_comm)
    wrk.load_data_books(job, otherargs, runargs, data)
    wrk.load_data_hdf5(job, otherargs, runargs, data)
    wrk.load_data_context(job, otherargs, runargs, data)

    # Save data
    wrk.save_data_books(job, otherargs, runargs, data)
    wrk.save_data_hdf5(job, otherargs, runargs, data)

    # Collect optional timing information
    if otherargs.timing_root is not None:
        alltimers = toast.timing.gather_timers(comm=toast_comm.comm_world)
        if toast_comm.world_rank == 0:
            toast.timing.dump(alltimers, otherargs.timing_root)

    log.info_rank("Conversion completed in", comm=comm, timer=timer)


def cli():
    world, procs, rank = toast.mpi.get_world()
    with toast.mpi.exception_guard(comm=world):
        main()


if __name__ == "__main__":
    cli()



# Copyright (c) 2018-2022 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simons Observatory HWP.

This module contains code for HWP analysis tools

"""
from .g3thwp import G3tHWP
from .hwp import (get_hwpss, subtract_hwpss, demod_tod, get_hwp_freq)
from .hwp_angle_model import apply_hwp_angle_model
from .sim_hwp import I_to_P_param
from .sim_hwp import sim_hwpss
from .sim_hwp import sim_hwpss_2f4f



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Mapmaking with the MADAM destriper.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops
from toast.observation import default_values as defaults

from .. import ops as so_ops
from .job import workflow_timer


def setup_mapmaker_madam(operators):
    """Add commandline args and operators for the MADAM mapmaker.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    if toast.ops.madam.available():
        operators.append(toast.ops.Madam(name="madam", enabled=False))


@workflow_timer
def mapmaker_madam(job, otherargs, runargs, data):
    """Run the MADAM mapmaker.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if toast.ops.madam.available() and job_ops.madam.enabled:
        job_ops.madam.params = toast.ops.madam_params_from_mapmaker(job_ops.mapmaker)
        job_ops.madam.pixel_pointing = job.pixels_final
        job_ops.madam.stokes_weights = job.weights_final
        job_ops.madam.apply(data)
